# encoding: utf-8

"""Unit test suite for `cr.cube.matrix.cubemeasure` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension, _ValidElements
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.matrix.cubemeasure import (
    _BaseCubeMeans,
    _BaseCubeMeasure,
    _BaseCubeOverlaps,
    _BaseCubeStdDev,
    _BaseCubeSums,
    BaseCubeResultMatrix,
    _BaseUnweightedCubeCounts,
    _BaseWeightedCubeCounts,
    _CatXCatCubeMeans,
    _CatXCatCubeStdDev,
    _CatXCatCubeSums,
    _CatXCatMatrix,
    _CatXCatUnweightedCubeCounts,
    _CatXCatWeightedCubeCounts,
    _CatXMrCubeMeans,
    _CatXMrCubeStdDev,
    _CatXMrCubeSums,
    _CatXMrMatrix,
    _CatXMrUnweightedCubeCounts,
    _CatXMrWeightedCubeCounts,
    CubeMeasures,
    _MrXCatCubeMeans,
    _MrXCatCubeStdDev,
    _MrXCatCubeSums,
    _MrXCatMatrix,
    _MrXCatUnweightedCubeCounts,
    _MrXCatWeightedCubeCounts,
    _MrXMrCubeMeans,
    _MrXMrCubeStdDev,
    _MrXMrCubeSums,
    _MrXMrMatrix,
    _MrXMrUnweightedCubeCounts,
    _MrXMrWeightedCubeCounts,
    _NumArrayXMrUnweightedCubeCounts,
    _NumArrayXCatUnweightedCubeCounts,
)

from ...unitutil import class_mock, instance_mock, method_mock, property_mock


class DescribeCubeMeasures(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure.CubeMeasures` object."""

    @pytest.mark.parametrize(
        "cube_measure_, CubeMeasureCls",
        (
            ("cube_means", _BaseCubeMeans),
            ("cube_overlaps", _BaseCubeOverlaps),
            ("cube_sum", _BaseCubeSums),
            ("cube_stddev", _BaseCubeStdDev),
        ),
    )
    def it_provides_access_to_cube_measures_object(
        self, request, cube_, dimensions_, cube_measure_, CubeMeasureCls
    ):
        _cube_measure_ = instance_mock(request, CubeMeasureCls)
        CubeMeasureCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeMeasureCls.__name__,
        )
        CubeMeasureCls_.factory.return_value = _cube_measure_
        cube_measures = CubeMeasures(cube_, dimensions_, slice_idx=37)

        cube_measure = getattr(cube_measures, cube_measure_)

        CubeMeasureCls_.factory.assert_called_once_with(cube_, dimensions_, 37)
        assert cube_measure is _cube_measure_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))


class Describe_BaseCubeMeasure(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeMeasure`."""

    @pytest.mark.parametrize(
        "ndim, tab_dim_type, expected_value",
        (
            (2, None, slice(None, None, None)),
            (3, DT.MR, (3, 0)),
            (3, DT.CAT, 3),
        ),
    )
    def it_computes_index_expression_that_selects_measure_values_from_cube_to_help(
        self, request, ndim, tab_dim_type, expected_value
    ):
        cube_ = instance_mock(request, Cube, ndim=ndim, dimension_types=(tab_dim_type,))
        measure = _BaseCubeMeasure(None)

        assert measure._slice_idx_expr(cube_, slice_idx=3) == expected_value


# === MEANS ===


class Describe_BaseCubeMeans(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeMeans`."""

    @pytest.mark.parametrize(
        "dimension_types, CubeMeansCls",
        (
            ((DT.MR, DT.MR), _MrXMrCubeMeans),
            ((DT.MR, DT.CAT), _MrXCatCubeMeans),
            ((DT.CAT, DT.MR), _CatXMrCubeMeans),
            ((DT.CAT, DT.CAT), _CatXCatCubeMeans),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_means_objects(
        self, request, dimension_types, CubeMeansCls
    ):
        cube_ = instance_mock(request, Cube)
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_means_ = instance_mock(request, CubeMeansCls)
        CubeMeanCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeMeansCls.__name__,
            return_value=cube_means_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeMeans,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.means = [[1, 2], [3, 4]]

        cube_means = _BaseCubeMeans.factory(cube_, dimensions_, slice_idx=7)

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeMeanCls_.assert_called_once_with(dimensions_, [3, 4])
        assert cube_means is cube_means_

    def but_it_raises_a_value_error_when_cube_result_does_not_contain_mean_measure(
        self, cube_
    ):
        cube_.means = None

        with pytest.raises(ValueError) as e:
            _BaseCubeMeans.factory(cube_, None, None)

        assert str(e.value) == "cube-result does not contain cube-means measure"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)


class Describe_CatXCatCubeMeans(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _CatXCatCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [1.1, 2.3, 3.3],
            [3.4, 1.5, 1.6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 3) np.float64 ndarray of means as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeMeans(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _CatXMrCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [1.1, 2.2, 3.2],
            [4.3, 5.1, 6.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 3, 2) np.float64 ndarray of means as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatCubeMeans(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _MrXCatCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [1.1, 6.1],
            [4.3, 3.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 3, 2) np.int float64 of means as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXMrCubeMeans(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _MrXMrCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [0.1, 0.1],
            [0.4, 0.5],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 2, 2, 2) np.float64 ndarray of means as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0.1, 0.8],  # -- col 0
                        [0.1, 0.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.2, 0.6],  # -- col 0
                        [0.3, 0.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [0.4, 0.4],  # -- col 0
                        [0.5, 0.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.6, 0.2],  # -- col 0
                        [0.7, 0.1],  # -- col 1
                    ],
                ],
            ]
        )


# === STD DEV ===


class Describe_BaseCubeStdDev(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeStdDev`."""

    @pytest.mark.parametrize(
        "dimension_types, CubeStdDevCls",
        (
            ((DT.MR, DT.MR), _MrXMrCubeStdDev),
            ((DT.MR, DT.CAT), _MrXCatCubeStdDev),
            ((DT.CAT, DT.MR), _CatXMrCubeStdDev),
            ((DT.CAT, DT.CAT), _CatXCatCubeStdDev),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_stddev_objects(
        self, request, dimension_types, CubeStdDevCls, cube_
    ):
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_stddev_ = instance_mock(request, CubeStdDevCls)
        CubeStdDevCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeStdDevCls.__name__,
            return_value=cube_stddev_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeStdDev,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.stddev = [[1, 2], [3, 4]]

        cube_stddev = _BaseCubeStdDev.factory(cube_, dimensions_, slice_idx=7)

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeStdDevCls_.assert_called_once_with(dimensions_, [3, 4])
        assert cube_stddev is cube_stddev_

    def but_it_raises_a_value_error_when_cube_result_does_not_contain_stddev_measure(
        self, cube_
    ):
        cube_.stddev = None

        with pytest.raises(ValueError) as e:
            _BaseCubeStdDev.factory(cube_, None, None)

        assert str(e.value) == "cube-result does not contain cube-stddev measure"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)


class Describe_CatXCatCubeStdDev(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _CatXCatCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [1.1, 2.3, 3.3],
            [3.4, 1.5, 1.6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 3) np.float64 ndarray of stddev as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeStdDev(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _CatXMrCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [1.1, 2.2, 3.2],
            [4.3, 5.1, 6.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 3, 2) np.float64 ndarray of stddev as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatCubeStdDev(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _MrXCatCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [1.1, 6.1],
            [4.3, 3.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 3, 2) np.int float64 of stddev as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXMrCubeStdDev(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _MrXMrCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [0.1, 0.1],
            [0.4, 0.5],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 2, 2, 2) np.float64 ndarray of stddev as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0.1, 0.8],  # -- col 0
                        [0.1, 0.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.2, 0.6],  # -- col 0
                        [0.3, 0.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [0.4, 0.4],  # -- col 0
                        [0.5, 0.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.6, 0.2],  # -- col 0
                        [0.7, 0.1],  # -- col 1
                    ],
                ],
            ]
        )


# === SUM ===


class Describe_BaseCubeSum(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeSum`."""

    @pytest.mark.parametrize(
        "dimension_types, CubeSumCls",
        (
            ((DT.MR, DT.MR), _MrXMrCubeSums),
            ((DT.MR, DT.CAT), _MrXCatCubeSums),
            ((DT.CAT, DT.MR), _CatXMrCubeSums),
            ((DT.CAT, DT.CAT), _CatXCatCubeSums),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_sum_objects(
        self, request, dimension_types, CubeSumCls, cube_
    ):
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_sums_ = instance_mock(request, CubeSumCls)
        CubeSumCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeSumCls.__name__,
            return_value=cube_sums_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeSums,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.sums = [[1, 2], [3, 4]]

        cube_sums = _BaseCubeSums.factory(cube_, dimensions_, slice_idx=7)

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeSumCls_.assert_called_once_with(dimensions_, [3, 4])
        assert cube_sums is cube_sums_

    def but_it_raises_a_value_error_when_cube_result_does_not_contain_sum_measure(
        self, cube_
    ):
        cube_.sums = None

        with pytest.raises(ValueError) as e:
            _BaseCubeSums.factory(cube_, None, None)

        assert str(e.value) == "cube-result does not contain cube-sum measure"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)


class Describe_CatXCatCubeSums(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeSums`."""

    def it_knows_its_sum(self, raw_sums):
        cube_sum = _CatXCatCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [1.1, 2.3, 3.3],
            [3.4, 1.5, 1.6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 3) np.float64 ndarray of sums as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeSum(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeSums`."""

    def it_knows_its_sum(self, raw_sums):
        cube_sum = _CatXMrCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [1.1, 2.2, 3.2],
            [4.3, 5.1, 6.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 3, 2) np.float64 ndarray of sums as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatCubeSum(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeSums`."""

    def it_knows_its_sum(self, raw_sums):
        cube_sum = _MrXCatCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [1.1, 6.1],
            [4.3, 3.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 3, 2) np.int float64 of sums as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXMrCubeSum(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeSums`."""

    def it_knows_its_sums(self, raw_sums):
        cube_sum = _MrXMrCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [0.1, 0.1],
            [0.4, 0.5],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 2, 2, 2) np.float64 ndarray of sums as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0.1, 0.8],  # -- col 0
                        [0.1, 0.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.2, 0.6],  # -- col 0
                        [0.3, 0.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [0.4, 0.4],  # -- col 0
                        [0.5, 0.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.6, 0.2],  # -- col 0
                        [0.7, 0.1],  # -- col 1
                    ],
                ],
            ]
        )


# === UNWEIGHTED COUNTS ===


class Describe_BaseUnweightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseUnweightedCubeCounts`."""

    @pytest.mark.parametrize(
        (
            "dimension_types",
            "UnweightedCubeCountsCls",
            "unweighted_counts",
            "unweighted_valid_counts",
            "diff_nans",
            "expected_counts",
        ),
        (
            (
                (DT.MR, DT.MR),
                _MrXMrUnweightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.MR, DT.CAT),
                _MrXCatUnweightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.CAT, DT.MR),
                _CatXMrUnweightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.CAT, DT.CAT),
                _CatXCatUnweightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.NUM_ARRAY, DT.MR),
                _NumArrayXMrUnweightedCubeCounts,
                [[1, 2], [3, 4]],
                [[1, 2], [6, 7]],
                True,
                [6, 7],
            ),
            (
                (DT.NUM_ARRAY, DT.CAT),
                _NumArrayXCatUnweightedCubeCounts,
                [[1, 2], [3, 4]],
                [[1, 2], [6, 7]],
                True,
                [6, 7],
            ),
        ),
    )
    def it_provides_a_factory_for_constructing_unweighted_cube_count_objects(
        self,
        request,
        dimension_types,
        UnweightedCubeCountsCls,
        unweighted_counts,
        unweighted_valid_counts,
        diff_nans,
        expected_counts,
    ):
        cube_ = instance_mock(request, Cube)
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        unweighted_cube_counts_ = instance_mock(request, UnweightedCubeCountsCls)
        UnweightedCubeCountsCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % UnweightedCubeCountsCls.__name__,
            return_value=unweighted_cube_counts_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseUnweightedCubeCounts,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.unweighted_valid_counts = unweighted_valid_counts
        cube_.unweighted_counts = unweighted_counts

        unweighted_cube_counts = _BaseUnweightedCubeCounts.factory(
            cube_, dimensions_, slice_idx=7
        )

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        UnweightedCubeCountsCls_.assert_called_once_with(
            dimensions_, expected_counts, diff_nans
        )
        assert unweighted_cube_counts is unweighted_cube_counts_

    @pytest.mark.parametrize(
        "columns_base, expected_value",
        (
            (np.arange(3), [[0, 1, 2], [0, 1, 2]]),
            (np.arange(6).reshape(2, 3), [[0, 1, 2], [3, 4, 5]]),
        ),
    )
    def it_computes_the_column_bases(self, request, columns_base, expected_value):
        property_mock(
            request,
            _BaseUnweightedCubeCounts,
            "columns_base",
            return_value=columns_base,
        )
        property_mock(
            request,
            _BaseUnweightedCubeCounts,
            "unweighted_counts",
            return_value=np.array([[0, 0, 0], [0, 0, 0]]),
        )
        unweighted_cube_counts = _BaseUnweightedCubeCounts(None, None, None)

        assert unweighted_cube_counts.column_bases.tolist() == expected_value


class Describe_CatXCatUnweightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatUnweightedCubeCounts`."""

    def it_knows_its_columns_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_base.tolist() == [5, 7, 9]

    def it_knows_its_columns_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_pruning_base.tolist() == [5, 7, 9]

    def it_knows_its_row_bases(self, request, raw_unweighted_counts):
        property_mock(
            request,
            _CatXCatUnweightedCubeCounts,
            "rows_base",
            return_value=np.array([2, 1]),
        )
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )

        assert unweighted_cube_counts.row_bases.tolist() == [[2, 2, 2], [1, 1, 1]]

    def it_knows_its_rows_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_pruning_base.tolist() == [6, 15]

    def it_knows_its_table_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.table_base == 21

    def it_knows_its_table_bases(self, request, raw_unweighted_counts):
        property_mock(
            request, _CatXCatUnweightedCubeCounts, "table_base", return_value=9
        )
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )

        assert unweighted_cube_counts.table_bases.tolist() == [[9, 9, 9], [9, 9, 9]]

    @pytest.mark.parametrize(
        ("raw_unweighted_counts", "expected"),
        (
            # --- (1, 3) array ---
            ([[1, 2, 3]], [6]),
            # --- (2, 3) array ---
            ([[1, 2, 3], [4, 5, 6]], [6, 15]),
            # --- (3, 1) array ---
            ([[1], [2], [3]], [1, 2, 3]),
        ),
    )
    def it_knows_its_rows_base(self, raw_unweighted_counts, expected):
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_base.tolist() == expected

    def it_knows_its_unweighted_counts(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.unweighted_counts.tolist() == [
            [1, 2, 3],
            [4, 5, 6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_unweighted_counts(self, request):
        """(2, 3) np.int ndarray of unweighted cube-counts as received from Cube."""
        return np.array([[1, 2, 3], [4, 5, 6]])


class Describe_CatXMrUnweightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrUnweightedCubeCounts`."""

    def it_knows_its_columns_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_base.tolist() == [5, 7, 9]

    def it_knows_its_columns_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_pruning_base.tolist() == [14, 14, 14]

    def it_knows_its_row_bases(self, request):
        property_mock(
            request,
            _CatXMrUnweightedCubeCounts,
            "rows_base",
            return_value=np.array([[1, 2, 3], [4, 5, 6]]),
        )
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(None, None, None)

        assert unweighted_cube_counts.row_bases.tolist() == [[1, 2, 3], [4, 5, 6]]

    def it_knows_its_rows_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_base.tolist() == [[7, 7, 7], [7, 7, 7]]

    def it_knows_its_rows_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_pruning_base.tolist() == [21, 21]

    def it_knows_its_table_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.table_base.tolist() == [14, 14, 14]

    def it_knows_its_table_bases(self, request, raw_unweighted_counts):
        property_mock(
            request,
            _CatXMrUnweightedCubeCounts,
            "table_base",
            return_value=np.array([5, 4, 3]),
        )
        property_mock(
            request,
            _CatXMrUnweightedCubeCounts,
            "unweighted_counts",
            return_value=np.array([[0, 0, 0], [0, 0, 0]]),
        )
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )

        assert unweighted_cube_counts.table_bases.tolist() == [[5, 4, 3], [5, 4, 3]]

    def it_knows_its_unweighted_counts(self, raw_unweighted_counts):
        unweighted_cube_counts = _CatXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.unweighted_counts.tolist() == [
            [1, 2, 3],
            [4, 5, 6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_unweighted_counts(self):
        """(2, 3, 2) np.int ndarray of unweighted cube-counts as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1, 6],  # -- col 0 --
                    [2, 5],  # -- col 1 --
                    [3, 4],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4, 3],  # -- col 0 --
                    [5, 2],  # -- col 1 --
                    [6, 1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatUnweightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatUnweightedCubeCounts`."""

    def it_knows_its_columns_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_base.tolist() == [[5, 7, 9], [7, 12, 11]]

    def it_knows_its_columns_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_pruning_base.tolist() == [12, 19, 20]

    def it_knows_its_row_bases(self, request):
        property_mock(
            request,
            _MrXCatUnweightedCubeCounts,
            "rows_base",
            return_value=np.array([1, 2, 3]),
        )
        property_mock(
            request,
            _MrXCatUnweightedCubeCounts,
            "unweighted_counts",
            return_value=np.array([[0, 0], [0, 0], [0, 0]]),
        )
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(None, None, None)

        assert unweighted_cube_counts.row_bases.tolist() == [[1, 1], [2, 2], [3, 3]]

    def it_knows_its_rows_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_base.tolist() == [6, 24]

    def it_knows_its_rows_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_pruning_base.tolist() == [21, 30]

    def it_knows_its_table_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.table_base.tolist() == [21, 30]

    def it_knows_its_table_bases(self, request, raw_unweighted_counts):
        property_mock(
            request,
            _MrXCatUnweightedCubeCounts,
            "table_base",
            return_value=np.array([6, 5]),
        )
        property_mock(
            request,
            _MrXCatUnweightedCubeCounts,
            "unweighted_counts",
            return_value=np.array([[0, 0, 0], [0, 0, 0]]),
        )
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )

        assert unweighted_cube_counts.table_bases.tolist() == [[6, 6, 6], [5, 5, 5]]

    def it_knows_its_unweighted_counts(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXCatUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.unweighted_counts.tolist() == [
            [1, 2, 3],
            [7, 8, 9],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_unweighted_counts(self):
        """(2, 3, 2) np.int ndarray of unweighted cube-counts as received from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols) --
            [
                # -- 0  1  2 -- cols ---
                [  # -- row 0 ----------
                    [1, 2, 3],  # -- sel
                    [4, 5, 6],  # -- not
                ],
                [  # -- row 1 ----------
                    [7, 8, 9],  # -- sel
                    [0, 4, 2],  # -- not
                ],
            ]
        )


class Describe_MrXMrUnweightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrUnweightedCubeCounts`."""

    def it_knows_its_columns_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_base.tolist() == [[2, 4], [10, 12]]

    def it_knows_its_columns_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.columns_pruning_base.tolist() == [12, 16]

    def it_knows_its_row_bases(self, request):
        property_mock(
            request,
            _MrXMrUnweightedCubeCounts,
            "rows_base",
            return_value=np.array([[1, 2], [5, 6]]),
        )
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(None, None, None)

        assert unweighted_cube_counts.row_bases.tolist() == [[1, 2], [5, 6]]

    def it_knows_its_rows_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_base.tolist() == [[8, 8], [8, 8]]

    def it_knows_its_rows_pruning_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.rows_pruning_base.tolist() == [16, 16]

    def it_knows_its_table_base(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.table_base.tolist() == [[16, 16], [16, 16]]

    def it_knows_its_table_bases(self, request):
        property_mock(
            request,
            _MrXMrUnweightedCubeCounts,
            "table_base",
            return_value=np.array([[3, 2], [7, 6]]),
        )
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(None, None, None)

        assert unweighted_cube_counts.table_bases.tolist() == [[3, 2], [7, 6]]

    def it_knows_its_unweighted_counts(self, raw_unweighted_counts):
        unweighted_cube_counts = _MrXMrUnweightedCubeCounts(
            None, raw_unweighted_counts, None
        )
        assert unweighted_cube_counts.unweighted_counts.tolist() == [[0, 1], [4, 5]]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_unweighted_counts(self):
        """(2, 2, 2, 2) np.int ndarray of unweighted cube-counts as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0, 8],  # -- col 0
                        [1, 7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [2, 6],  # -- col 0
                        [3, 5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [4, 4],  # -- col 0
                        [5, 3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [6, 2],  # -- col 0
                        [7, 1],  # -- col 1
                    ],
                ],
            ]
        )


# === WEIGHTED COUNTS ===


class Describe_BaseWeightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseWeightedCubeCounts`."""

    @pytest.mark.parametrize(
        (
            "dimension_types",
            "WeightedCubeCountsCls",
            "weighted_counts",
            "weighted_valid_counts",
            "diff_nans",
            "expected_counts",
        ),
        (
            (
                (DT.MR, DT.MR),
                _MrXMrWeightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.MR, DT.CAT),
                _MrXCatWeightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.CAT, DT.MR),
                _CatXMrWeightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.CAT, DT.CAT),
                _CatXCatWeightedCubeCounts,
                [[1, 2], [3, 4]],
                None,
                False,
                [3, 4],
            ),
            (
                (DT.CAT, DT.CAT),
                _CatXCatWeightedCubeCounts,
                [[1, 2], [3, 4]],
                [[1, 2], [5, 4]],
                True,
                [5, 4],
            ),
        ),
    )
    def it_provides_a_factory_for_constructing_weighted_cube_count_objects(
        self,
        request,
        cube_,
        dimensions_,
        dimension_types,
        WeightedCubeCountsCls,
        weighted_counts,
        weighted_valid_counts,
        diff_nans,
        expected_counts,
    ):
        weighted_cube_counts_ = instance_mock(request, WeightedCubeCountsCls)
        WeightedCubeCountsCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % WeightedCubeCountsCls.__name__,
            return_value=weighted_cube_counts_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseWeightedCubeCounts,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.counts = weighted_counts
        cube_.weighted_valid_counts = weighted_valid_counts

        weighted_cube_counts = _BaseWeightedCubeCounts.factory(
            cube_, dimensions_, slice_idx=2
        )

        _slice_idx_expr_.assert_called_once_with(cube_, 2)
        WeightedCubeCountsCls_.assert_called_once_with(
            dimensions_, expected_counts, diff_nans
        )
        assert weighted_cube_counts is weighted_cube_counts_

    @pytest.mark.parametrize(
        "columns_margin, expected_value",
        (
            (7.7, [[7.7, 7.7], [7.7, 7.7], [7.7, 7.7]]),
            (np.array([1.1, 2.2]), [[1.1, 2.2], [1.1, 2.2], [1.1, 2.2]]),
            (
                np.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5]]),
                [[0.0, 1.1], [2.2, 3.3], [4.4, 5.5]],
            ),
        ),
    )
    def it_computes_the_column_bases(self, request, columns_margin, expected_value):
        property_mock(
            request,
            _BaseWeightedCubeCounts,
            "columns_margin",
            return_value=columns_margin,
        )
        property_mock(
            request,
            _BaseWeightedCubeCounts,
            "weighted_counts",
            return_value=np.array([[0, 0], [0, 0], [0, 0]]),
        )
        weighted_cube_counts = _BaseWeightedCubeCounts(None, None, False)

        assert weighted_cube_counts.column_bases.tolist() == expected_value

    def it_can_compute_array_type_std_residual_to_help(self):
        counts = np.array([[0, 2, 4, 6], [8, 10, 12, 14], [16, 18, 20, 22]])
        total = np.array([51, 63, 75, 87])
        rowsum = np.array([[1, 5, 9, 13], [17, 21, 25, 29], [33, 37, 41, 45]])
        colsum = np.array([24, 30, 36, 42])
        matrix = _BaseWeightedCubeCounts(None, None, False)

        residuals = matrix._array_type_std_res(counts, total, rowsum, colsum)

        assert pytest.approx(residuals) == [
            [-0.9521905, -0.3555207, -0.2275962, -0.1660169],
            [0.0, 0.0, 0.0, 0.0],
            [0.2762585, 0.1951985, 0.1485686, 0.1184429],
        ]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimensions_(self, request):
        return instance_mock(request, Dimension), instance_mock(request, Dimension)


class Describe_CatXCatWeightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatWeightedCubeCounts`."""

    def it_knows_its_columns_margin(self, raw_weighted_counts):
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )

        assert weighted_cube_counts.columns_margin == pytest.approx(
            np.array([9.9, 7.7, 5.5])
        )

    def it_knows_its_row_bases(self, request, raw_weighted_counts):
        property_mock(
            request,
            _CatXCatWeightedCubeCounts,
            "rows_margin",
            return_value=np.array([8.8, 9.9]),
        )
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )

        assert weighted_cube_counts.row_bases == pytest.approx(
            np.array([[8.8, 8.8, 8.8], [9.9, 9.9, 9.9]])
        )

    def it_knows_its_rows_margin(self, raw_weighted_counts):
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.rows_margin.tolist() == [6.6, 16.5]

    def it_knows_its_table_bases(self, request, raw_weighted_counts):
        property_mock(
            request, _CatXCatWeightedCubeCounts, "table_margin", return_value=9.9
        )
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )

        assert weighted_cube_counts.table_bases.tolist() == [
            [9.9, 9.9, 9.9],
            [9.9, 9.9, 9.9],
        ]

    def it_knows_its_table_margin(self, raw_weighted_counts):
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.table_margin == 23.1

    def it_knows_its_weighted_counts(self, raw_weighted_counts):
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.weighted_counts.tolist() == [
            [3.3, 2.2, 1.1],
            [6.6, 5.5, 4.4],
        ]

    def it_knows_its_zscores(self, raw_weighted_counts):
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.zscores.tolist() == [
            [0.43874821936960656, 4.3387889766250713e-16, -0.5097793640389925],
            [-0.4387482193696045, 8.677577953250143e-16, 0.5097793640389934],
        ]

    def but_its_zscores_are_NaNs_for_a_deficient_matrix(self):
        raw_weighted_counts = np.array([[1.1, 1.1], [2.2, 2.2]])
        weighted_cube_counts = _CatXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.zscores == pytest.approx(
            np.array([[np.nan, np.nan], [np.nan, np.nan]]), nan_ok=True
        )

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_weighted_counts(self):
        """(2, 3) np.float64 ndarray of weighted cube-counts as received from Cube."""
        return np.array([[3.3, 2.2, 1.1], [6.6, 5.5, 4.4]])


class Describe_CatXMrWeightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrWeightedCubeCounts`."""

    def it_knows_its_columns_margin(self, raw_weighted_counts):
        weighted_cube_counts = _CatXMrWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.columns_margin == pytest.approx(
            np.array([5.5, 7.7, 9.9])
        )

    def it_knows_its_row_bases(self, request):
        property_mock(
            request,
            _CatXMrWeightedCubeCounts,
            "rows_margin",
            return_value=np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]),
        )
        weighted_cube_counts = _CatXMrWeightedCubeCounts(None, None, None)

        assert weighted_cube_counts.row_bases.tolist() == [
            [1.1, 2.2, 3.3],
            [4.4, 5.5, 6.6],
        ]

    def it_knows_its_rows_margin(self, raw_weighted_counts):
        weighted_cube_counts = _CatXMrWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.rows_margin == pytest.approx(
            np.array([[7.7, 7.7, 7.7], [7.7, 7.7, 7.7]])
        )

    def it_knows_its_table_bases(self, request):
        property_mock(
            request,
            _CatXMrWeightedCubeCounts,
            "table_margin",
            return_value=np.array([9.8, 7.6, 5.4]),
        )
        property_mock(
            request,
            _CatXMrWeightedCubeCounts,
            "weighted_counts",
            return_value=np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
        )
        weighted_cube_counts = _CatXMrWeightedCubeCounts(None, None, None)

        assert weighted_cube_counts.table_bases.tolist() == [
            [9.8, 7.6, 5.4],
            [9.8, 7.6, 5.4],
        ]

    def it_knows_its_table_margin(self, raw_weighted_counts):
        weighted_cube_counts = _CatXMrWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.table_margin == pytest.approx([15.4, 15.4, 15.4])

    def it_knows_its_weighted_counts(self, raw_weighted_counts):
        weighted_cube_counts = _CatXMrWeightedCubeCounts(
            None, raw_weighted_counts, None
        )

        assert weighted_cube_counts.weighted_counts == pytest.approx(
            np.array(
                [
                    [1.1, 2.2, 3.3],
                    [4.4, 5.5, 6.6],
                ]
            )
        )

    def it_knows_its_zscores(self, raw_weighted_counts):
        weighted_cube_counts = _CatXMrWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert pytest.approx(weighted_cube_counts.zscores) == [
            [-1.7549928774784245, -1.6818357317441646, -1.754992877478425],
            [1.7549928774784245, 1.6818357317441637, 1.754992877478425],
        ]

    def but_its_zscores_are_NaNs_for_a_deficient_matrix(self):
        raw_weighted_counts = np.array([[[1.2, 1.2], [2.1, 2.2]]])
        weighted_cube_counts = _CatXMrWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.zscores == pytest.approx(
            np.array([[np.nan, np.nan]]), nan_ok=True
        )

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_weighted_counts(self):
        """(2, 3, 2) np.float ndarray of weighted cube-counts as received from Cube."""
        return np.array(
            [
                # -- sel / not ---------------
                [  # -- row 0 ----------------
                    [1.1, 6.6],  # -- col 0 --
                    [2.2, 5.5],  # -- col 1 --
                    [3.3, 4.4],  # -- col 2 --
                ],
                [  # -- row 1 ----------------
                    [4.4, 3.3],  # -- col 0 --
                    [5.5, 2.2],  # -- col 1 --
                    [6.6, 1.1],  # -- col 2 --
                ],
            ]
        )


class Describe_MrXCatWeightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatWeightedCubeCounts`."""

    def it_knows_its_columns_margin(self, raw_weighted_counts):
        weighted_cube_counts = _MrXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.columns_margin == pytest.approx(
            np.array([[5.5, 7.7, 9.9], [7.7, 13.2, 12.1]])
        )

    def it_knows_its_row_bases(self, request):
        property_mock(
            request,
            _MrXCatWeightedCubeCounts,
            "rows_margin",
            return_value=np.array([1.1, 2.2, 3.3]),
        )
        property_mock(
            request,
            _MrXCatWeightedCubeCounts,
            "weighted_counts",
            return_value=np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]]),
        )
        weighted_cube_counts = _MrXCatWeightedCubeCounts(None, None, None)

        assert weighted_cube_counts.row_bases == pytest.approx(
            np.array([[1.1, 1.1], [2.2, 2.2], [3.3, 3.3]])
        )

    def it_knows_its_rows_margin(self, raw_weighted_counts):
        weighted_cube_counts = _MrXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.rows_margin == pytest.approx(np.array([6.6, 26.4]))

    def it_knows_its_table_bases(self, request):
        property_mock(
            request,
            _MrXCatWeightedCubeCounts,
            "table_margin",
            return_value=np.array([9.8, 7.6]),
        )
        property_mock(
            request,
            _MrXCatWeightedCubeCounts,
            "weighted_counts",
            return_value=np.arange(6).reshape(2, 3),
        )
        weighted_cube_counts = _MrXCatWeightedCubeCounts(None, None, None)

        assert weighted_cube_counts.table_bases.tolist() == [
            [9.8, 9.8, 9.8],
            [7.6, 7.6, 7.6],
        ]

    def it_knows_its_table_margin(self, raw_weighted_counts):
        weighted_cube_counts = _MrXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.table_margin == pytest.approx(
            np.array([23.1, 33.0])
        )

    def it_knows_its_weighted_counts(self, raw_weighted_counts):
        weighted_cube_counts = _MrXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.weighted_counts.tolist() == pytest.approx(
            np.array([[1.1, 2.2, 3.3], [7.7, 8.8, 9.9]])
        )

    def it_knows_its_zscores(self, raw_weighted_counts):
        weighted_cube_counts = _MrXCatWeightedCubeCounts(
            None, raw_weighted_counts, None
        )
        assert weighted_cube_counts.zscores.tolist() == [
            [-0.5097793640389925, 4.3387889766250713e-16, 0.43874821936960656],
            [1.584572360360167, -1.5634719199411429, 0.19867985355975712],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_weighted_counts(self):
        """(2, 2, 3) np.float ndarray of weighted cube-counts as received from Cube."""
        return np.array(
            [  # --   0    1    2  cols ------
                [  # -- row 0 ----------------
                    [1.1, 2.2, 3.3],  # -- sel
                    [4.4, 5.5, 6.6],  # -- not
                ],
                [  # -- row 1 ----------------
                    [7.7, 8.8, 9.9],  # -- sel
                    [0.0, 4.4, 2.2],  # -- not
                ],
            ]
        )


class Describe_MrXMrWeightedCubeCounts(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrWeightedCubeCounts`."""

    def it_knows_its_columns_margin(self, raw_weighted_counts):
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, raw_weighted_counts, None)
        assert weighted_cube_counts.columns_margin == pytest.approx(
            np.array([[2.2, 4.4], [11.0, 13.2]])
        )

    def it_knows_its_row_bases(self, request):
        property_mock(
            request,
            _MrXMrWeightedCubeCounts,
            "rows_margin",
            return_value=np.array([[1.1, 2.2], [5.5, 6.6]]),
        )
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, None, None)

        assert weighted_cube_counts.row_bases == pytest.approx(
            np.array([[1.1, 2.2], [5.5, 6.6]])
        )

    def it_knows_its_rows_margin(self, raw_weighted_counts):
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, raw_weighted_counts, None)
        assert weighted_cube_counts.rows_margin == pytest.approx(
            np.array([[8.8, 8.8], [8.8, 8.8]])
        )

    def it_knows_its_table_bases(self, request):
        property_mock(
            request,
            _MrXMrWeightedCubeCounts,
            "table_margin",
            return_value=np.array([[3.3, 2.2, 1.1], [7.7, 6.6, 5.5]]),
        )
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, None, None)

        assert weighted_cube_counts.table_bases == pytest.approx(
            np.array([[3.3, 2.2, 1.1], [7.7, 6.6, 5.5]])
        )

    def it_knows_its_table_margin(self, raw_weighted_counts):
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, raw_weighted_counts, None)
        assert weighted_cube_counts.table_margin == pytest.approx(
            np.array([[17.6, 17.6], [17.6, 17.6]])
        )

    def it_knows_its_weighted_counts(self, raw_weighted_counts):
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, raw_weighted_counts, None)

        assert weighted_cube_counts.weighted_counts == pytest.approx(
            np.array([[0.0, 1.1], [4.4, 5.5]])
        )

    def it_knows_its_zscores(self, raw_weighted_counts):
        weighted_cube_counts = _MrXMrWeightedCubeCounts(None, raw_weighted_counts, None)
        assert weighted_cube_counts.zscores.tolist() == [
            [-1.5856499343441839, -1.2110601416389966],
            [-1.0832051206181277, -1.2110601416389961],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_weighted_counts(self):
        """(2, 2, 2, 2) np.float ndarray of weighted cube-counts as from Cube."""
        return np.array(
            [  # ------ sel/not (col) -----
                [  # -- row 0 -------------
                    [  # -- selected ------
                        [0.0, 8.8],  # -- col 0
                        [1.1, 7.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [2.2, 6.6],  # -- col 0
                        [3.3, 5.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [4.4, 4.4],  # -- col 0
                        [5.5, 3.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [6.6, 2.2],  # -- col 0
                        [7.7, 1.1],  # -- col 1
                    ],
                ],
            ]
        )


# === LEGACY CUBE-RESULT MATRIX TESTS (should go away after measure consolidation) ===


class DescribeBaseCubeResultMatrix(object):
    """Unit test suite for `cr.cube.matrix.cubemeasure.BaseCubeResultMatrix` object."""

    def it_knows_its_column_proportions(self, request):
        property_mock(
            request,
            BaseCubeResultMatrix,
            "weighted_counts",
            return_value=np.array([[1, 2, 3], [4, 5, 6]]),
        )
        property_mock(
            request,
            BaseCubeResultMatrix,
            "columns_margin",
            return_value=np.array([5, 7, 9]),
        )
        matrix = BaseCubeResultMatrix(None, None, None)

        np.testing.assert_almost_equal(
            matrix.column_proportions,
            np.array([[0.2, 0.2857143, 0.3333333], [0.8, 0.7142857, 0.6666667]]),
        )

    def it_knows_its_columns_dimension(self, dimension_):
        matrix = BaseCubeResultMatrix([None, dimension_], None, None)
        assert matrix.columns_dimension == dimension_

    def it_knows_its_rows_dimension(self, dimension_):
        matrix = BaseCubeResultMatrix([dimension_, None], None, None)
        assert matrix.rows_dimension == dimension_

    @pytest.mark.parametrize(
        "dimension_types, MatrixCls",
        (
            ((DT.MR, DT.MR), _MrXMrMatrix),
            ((DT.MR, DT.CAT), _MrXCatMatrix),
            ((DT.CAT, DT.MR), _CatXMrMatrix),
            ((DT.CAT, DT.CAT), _CatXCatMatrix),
        ),
    )
    def it_provides_a_factory_for_constructing_a_matrix_objects(
        self, request, dimension_types, MatrixCls
    ):
        cube_ = instance_mock(request, Cube)
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_.dimension_types = dimension_types
        MatrixCls_ = class_mock(
            request, "cr.cube.matrix.cubemeasure.%s" % MatrixCls.__name__
        )

        _sliced_counts = method_mock(
            request,
            BaseCubeResultMatrix,
            "_sliced_counts",
            return_value=([[1], [2]], [[3], [4]]),
        )

        matrix = BaseCubeResultMatrix.factory(cube_, dimensions_, slice_idx=17)

        _sliced_counts.assert_called_once_with(cube_, 17)
        MatrixCls_.assert_called_once_with(dimensions_, [[1], [2]], [[3], [4]])
        assert matrix is MatrixCls_.return_value

    @pytest.mark.parametrize(
        ("slice_idx", "dim_types", "expected"),
        (
            # --- <= 2D ---
            (0, (DT.CAT,), np.s_[:]),
            (1, (DT.CAT,), np.s_[:]),
            (0, (DT.CAT, DT.CAT), np.s_[:]),
            (1, (DT.CAT, DT.CAT), np.s_[:]),
            # --- 3D, no MR as tabs ---
            (0, (DT.CAT, DT.CAT, DT.CAT), np.s_[0]),
            (1, (DT.CAT, DT.CAT, DT.CAT), np.s_[1]),
            (2, (DT.CAT, DT.CAT, DT.CAT), np.s_[2]),
            # --- 3D, MR as tabs ---
            (0, (DT.MR, DT.CAT, DT.CAT), np.s_[0, 0]),
            (1, (DT.MR, DT.CAT, DT.CAT), np.s_[1, 0]),
            (2, (DT.MR, DT.CAT, DT.CAT), np.s_[2, 0]),
        ),
    )
    def it_knows_its_cube_slice_expression_to_help(
        self, cube_, slice_idx, dim_types, expected
    ):
        cube_.dimension_types = dim_types
        cube_.ndim = len(dim_types)

        s = BaseCubeResultMatrix._cube_slice_expression(cube_, slice_idx)

        assert s == expected

    @pytest.mark.parametrize(
        ("counts", "counts_slice", "expected"),
        (
            ([[1, 2, 3], [4, 5, 6]], np.s_[:], [[1, 2, 3], [4, 5, 6]]),
            ([[1, 2, 3], [4, 5, 6]], np.s_[0], [1, 2, 3]),
            ([[1, 2, 3], [4, 5, 6]], np.s_[0, 0], 1),
        ),
    )
    def it_knows_its_sliced_counts_to_help(
        self, request, cube_, counts, counts_slice, expected
    ):
        counts = np.array(counts)
        cube_.counts = counts
        cube_.unweighted_counts = counts
        cube_.counts_with_missings = counts
        _cube_slice_expression_ = method_mock(
            request,
            BaseCubeResultMatrix,
            "_cube_slice_expression",
            return_value=counts_slice,
        )

        sliced_counts = BaseCubeResultMatrix._sliced_counts(cube_, slice_idx=23)

        _cube_slice_expression_.assert_called_once_with(cube_, 23)
        counts, unweighted, with_missing = sliced_counts
        assert counts.tolist() == expected
        assert unweighted.tolist() == expected

    def it_produces_a_valid_row_indexer_to_help(self, request, dimension_):
        dimension_.valid_elements = instance_mock(
            request, _ValidElements, element_idxs=(0, 1, 2)
        )
        matrix = BaseCubeResultMatrix((dimension_, None), None, None)

        valid_row_idxs = matrix._valid_row_idxs

        assert isinstance(valid_row_idxs, tuple)
        assert len(valid_row_idxs) == 1
        assert valid_row_idxs[0].tolist() == [0, 1, 2]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_CatXCatMatrix(object):
    """Unit test suite for `cr.cube.matrix._CatXCatMatrix` object."""

    def it_knows_its_columns_index(self, request):
        property_mock(
            request,
            _CatXCatMatrix,
            "column_proportions",
            return_value=np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
        )
        property_mock(
            request,
            _CatXCatMatrix,
            "_baseline",
            return_value=np.array([[0.2], [0.8]]),
        )
        matrix = _CatXCatMatrix(None, None, None)

        np.testing.assert_almost_equal(
            matrix.column_index, np.array([[50.0, 100.0, 150.0], [50.0, 62.5, 75.0]])
        )

    def it_knows_its_columns_margin(self):
        matrix = _CatXCatMatrix(None, np.array([[1, 2, 3], [4, 5, 6]]), None)
        assert matrix.columns_margin.tolist() == [5, 7, 9]

    @pytest.mark.parametrize(
        ("unweighted_counts", "expected"),
        (
            # --- 1 "row", 0 cols ---
            ([[]], []),
            # --- 1 row, 1 col ---
            ([[1]], [1]),
            # --- 1 row, 3 cols ---
            ([[1, 2, 3]], [1, 2, 3]),
            # --- 3 rows, 0 cols ---
            ([[], [], []], []),
            # --- 3 rows, 1 col ---
            ([[1], [2], [3]], [6]),
            # --- 3 rows, 3 cols ---
            ([[1, 2, 3], [4, 5, 6]], [5, 7, 9]),
        ),
    )
    def it_knows_its_columns_pruning_base(self, unweighted_counts, expected):
        matrix = _CatXCatMatrix(None, None, np.array(unweighted_counts))

        columns_pruning_base = matrix.columns_pruning_base

        assert columns_pruning_base.shape == (len(expected),)
        assert columns_pruning_base.tolist() == expected

    @pytest.mark.parametrize(
        ("unweighted_counts", "expected"),
        (
            ([[1, 2, 3]], [6]),
            ([[1, 2, 3], [4, 5, 6]], [6, 15]),
            ([[1], [2], [3]], [1, 2, 3]),
        ),
    )
    def it_knows_its_rows_base(self, unweighted_counts, expected):
        matrix = _CatXCatMatrix(None, None, unweighted_counts)
        assert matrix.rows_base.tolist() == expected

    def it_knows_its_rows_margin(self):
        weighted_counts = np.array([[1, 2, 3], [4, 5, 6]])
        matrix = _CatXCatMatrix(None, weighted_counts, None)

        assert matrix.rows_margin.tolist() == [6, 15]

    @pytest.mark.parametrize(
        ("unweighted_counts", "expected"),
        (
            ([[1, 2, 3]], [6]),
            ([[1, 2, 3], [4, 5, 6]], [6, 15]),
            ([[1], [2], [3]], [1, 2, 3]),
        ),
    )
    def it_knows_its_rows_pruning_base(self, unweighted_counts, expected):
        matrix = _CatXCatMatrix(None, None, unweighted_counts)
        assert matrix.rows_pruning_base.tolist() == expected

    def it_knows_its_table_base(self):
        unweighted_counts = np.array([[1, 2, 3], [4, 5, 6]])
        assert _CatXCatMatrix(None, None, unweighted_counts).table_base == 21

    def it_knows_its_table_margin(self):
        weighted_counts = np.array([[1, 2, 3], [4, 5, 6]])
        assert _CatXCatMatrix(None, weighted_counts, None).table_margin == 21

    def it_knows_its_table_stderrs(self, request):
        property_mock(
            request,
            _CatXCatMatrix,
            "_table_proportion_variances",
            return_value=np.array([[0.5, 1.0], [0.25, 2.0]]),
        )
        property_mock(request, _CatXCatMatrix, "table_margin", return_value=2.0)

        np.testing.assert_almost_equal(
            _CatXCatMatrix(None, None, None).table_stderrs,
            np.array([[0.5, 0.7071068], [0.3535534, 1.0]]),
        )

    def it_knows_its_unweighted_counts(self):
        unweighted_cube_counts = np.array([[1, 2, 3], [4, 5, 6]])
        matrix = _CatXCatMatrix(None, None, unweighted_cube_counts)

        assert matrix.unweighted_counts.tolist() == [[1, 2, 3], [4, 5, 6]]

    def it_knows_its_weighted_counts(self):
        weighted_cube_counts = np.array([[3, 2, 1], [6, 5, 4]])
        matrix = _CatXCatMatrix(None, weighted_cube_counts, None)

        assert matrix.weighted_counts.tolist() == [[3, 2, 1], [6, 5, 4]]

    def it_knows_its_baseline_to_help(self, request):
        property_mock(
            request, _CatXCatMatrix, "_valid_row_idxs", return_value=np.array([0, 1])
        )
        counts_with_missings = np.array([[1, 2, 3], [4, 5, 6]])

        np.testing.assert_almost_equal(
            _CatXCatMatrix(None, None, None, counts_with_missings)._baseline,
            np.array([[0.2857143], [0.7142857]]),
        )

    def it_knows_its_table_proportion_variances_to_help(self, request):
        weighted_counts = np.arange(6).reshape(2, 3)
        np.testing.assert_almost_equal(
            _CatXCatMatrix(None, weighted_counts, None)._table_proportion_variances,
            np.array([[0.0, 0.0622222, 0.1155556], [0.16, 0.1955556, 0.2222222]]),
        )


class Describe_CatXMrMatrix(object):
    """Unit test suite for `cr.cube.matrix._CatXMrMatrix` object."""

    def it_knows_its_columns_pruning_base(self):
        unweighted_cube_counts = np.array(
            [
                [[1, 6], [2, 5], [3, 5]],  # --- row 0 ---
                [[5, 3], [6, 3], [7, 2]],  # --- row 1 ---
            ]
        )
        matrix = _CatXMrMatrix(None, None, unweighted_cube_counts, None)

        assert matrix.columns_pruning_base.tolist() == [15, 16, 17]

    def it_knows_its_rows_base(self):
        unweighted_cube_counts = np.array(
            [[[1, 6], [2, 5], [3, 4]], [[5, 3], [6, 2], [7, 1]]]
        )
        matrix = _CatXMrMatrix(None, None, unweighted_cube_counts, None)

        assert matrix.rows_base.tolist() == [[7, 7, 7], [8, 8, 8]]

    def it_knows_its_rows_margin(self):
        weighted_cube_counts = np.array(
            [[[1, 6], [2, 5], [3, 4]], [[5, 3], [6, 2], [7, 1]]]
        )
        matrix = _CatXMrMatrix(None, weighted_cube_counts, None, None)

        assert matrix.rows_margin.tolist() == [[7, 7, 7], [8, 8, 8]]

    def it_knows_its_rows_pruning_base(self):
        unweighted_cube_counts = np.array(
            [[[1, 6], [2, 5], [3, 4]], [[5, 3], [6, 2], [7, 1]]]
        )
        matrix = _CatXMrMatrix(None, None, unweighted_cube_counts, None)

        assert matrix.rows_pruning_base.tolist() == [21, 24]

    def it_knows_its_table_base(self):
        unweighted_cube_counts = np.array(
            [
                [  # -- row 0 ------------
                    [1, 6],  # -- col 0 --
                    [2, 5],  # -- col 1 --
                    [3, 4],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4, 3],  # -- col 0 --
                    [5, 2],  # -- col 1 --
                    [6, 1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )
        matrix = _CatXMrMatrix(None, None, unweighted_cube_counts, None)

        assert matrix.table_base.tolist() == [14, 14, 14]

    def it_knows_its_table_margin(self):
        weighted_cube_counts = np.array(
            [
                [  # -- row 0 ------------
                    [1, 6],  # -- col 0 --
                    [2, 5],  # -- col 1 --
                    [3, 4],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4, 3],  # -- col 0 --
                    [5, 2],  # -- col 1 --
                    [6, 1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )
        matrix = _CatXMrMatrix(None, weighted_cube_counts, None, None)

        assert matrix.table_margin.tolist() == [14, 14, 14]

    def it_knows_its_unweighted_counts(self):
        unweighted_cube_counts = np.array(
            [
                [  # -- row 0 ------------
                    [1, 6],  # -- col 0 --
                    [2, 5],  # -- col 1 --
                    [3, 4],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4, 3],  # -- col 0 --
                    [5, 2],  # -- col 1 --
                    [6, 1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )
        matrix = _CatXMrMatrix(None, None, unweighted_cube_counts, None)

        assert matrix.unweighted_counts.tolist() == [[1, 2, 3], [4, 5, 6]]

    def it_knows_its_weighted_counts(self):
        weighted_cube_counts = np.array(
            [
                [  # -- row 0 ------------
                    [1, 6],  # -- col 0 --
                    [2, 5],  # -- col 1 --
                    [3, 4],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4, 3],  # -- col 0 --
                    [5, 2],  # -- col 1 --
                    [6, 1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )
        matrix = _CatXMrMatrix(None, weighted_cube_counts, None, None)

        assert matrix.weighted_counts.tolist() == [[1, 2, 3], [4, 5, 6]]

    def it_knows_its_baseline_to_help(self, request):
        property_mock(
            request, _CatXMrMatrix, "_valid_row_idxs", return_value=np.array([0, 1])
        )
        counts_with_missings = np.array(
            [[[1, 6], [2, 5], [3, 4]], [[4, 3], [5, 2], [6, 1]]]
        )

        np.testing.assert_almost_equal(
            _CatXMrMatrix(None, None, None, counts_with_missings)._baseline,
            np.array([[0.5, 0.5, 0.5], [0.5, 0.5, 0.5]]),
        )

    def it_knows_its_table_proportion_variances_to_help(self, request):
        weighted_cube_counts = np.arange(12).reshape((2, 3, 2))
        np.testing.assert_almost_equal(
            _CatXMrMatrix(None, weighted_cube_counts, None)._table_proportion_variances,
            np.array([[0.0, 0.0826446, 0.1155556], [0.244898, 0.231405, 0.2222222]]),
        )


class Describe_MrXCatMatrix(object):
    """Unit test suite for `cr.cube.matrix._MrXCatMatrix` object."""

    def it_knows_its_columns_margin(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [1, 2, 3],  # -- selected
                    [4, 5, 6],  # -- not
                ],
                [  # -- row 1 ---------------
                    [7, 8, 9],  # -- selected
                    [3, 2, 1],  # -- not
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, weighted_counts, None, None).columns_margin,
            np.array([[5, 7, 9], [10, 10, 10]]),
        )

    def it_knows_its_columns_pruning_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [1, 2, 3],  # -- selected
                    [4, 5, 6],  # -- not
                ],
                [  # -- row 1 ---------------
                    [7, 8, 9],  # -- selected
                    [3, 2, 1],  # -- not
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, None, unweighted_counts, None).columns_pruning_base,
            np.array([15, 17, 19]),
        )

    def it_knows_its_rows_pruning_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ------------
                    [1, 2, 3],  # -- selected --
                    [4, 5, 6],  # -- not --
                ],
                [  # -- row 1 ------------
                    [7, 8, 9],  # -- selected --
                    [3, 2, 1],  # -- not --
                    # --------------------
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, None, unweighted_counts, None).rows_pruning_base,
            np.array([21, 30]),
        )

    def it_knows_its_table_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [1, 2, 3],  # -- selected
                    [4, 5, 6],  # -- not
                ],
                [  # -- row 1 ---------------
                    [7, 8, 9],  # -- selected
                    [0, 4, 2],  # -- not
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, None, unweighted_counts).table_base, [21, 30]
        )

    def it_knows_its_table_margin(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [1, 2, 3],  # -- selected
                    [4, 5, 6],  # -- not
                ],
                [  # -- row 1 ---------------
                    [7, 8, 9],  # -- selected
                    [0, 4, 2],  # -- not
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, weighted_counts, None).table_margin, [21, 30]
        )

    def it_knows_its_table_stderrs(self, request):
        property_mock(
            request,
            _MrXCatMatrix,
            "_table_proportion_variances",
            return_value=np.array([[0.5, 1.0], [0.25, 2.0]]),
        )
        property_mock(
            request,
            _MrXCatMatrix,
            "table_margin",
            return_value=np.array([[1, 2], [3, 4]]),
        )

        np.testing.assert_almost_equal(
            _MrXCatMatrix(None, None, None).table_stderrs,
            np.array(
                [
                    [[0.7071068, 0.7071068], [0.5, 1.0]],
                    [[0.4082483, 0.5], [0.2886751, 0.7071068]],
                ]
            ),
        )

    def it_knows_its_unweighted_counts(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ------------
                    [1, 2, 3],  # -- selected --
                    [4, 5, 6],  # -- not --
                ],
                [  # -- row 1 ------------
                    [7, 8, 9],  # -- selected --
                    [0, 4, 2],  # -- not --
                    # --------------------
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, None, unweighted_counts).unweighted_counts,
            np.array([[1, 2, 3], [7, 8, 9]]),
        )

    def it_knows_its_weighted_counts(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [1, 2, 3],  # -- selected
                    [4, 5, 6],  # -- not
                ],
                [  # -- row 1 ---------------
                    [7, 8, 9],  # -- selected
                    [0, 4, 2],  # -- not
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXCatMatrix(None, weighted_counts, None).weighted_counts,
            np.array([[1, 2, 3], [7, 8, 9]]),
        )

    def it_knows_its_baseline_to_help(self, request):
        property_mock(
            request, _MrXCatMatrix, "_valid_row_idxs", return_value=np.array([0, 1])
        )
        counts_with_missings = np.array(
            [[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [0, 4, 2]]]
        )

        np.testing.assert_almost_equal(
            _MrXCatMatrix(None, None, None, counts_with_missings)._baseline,
            np.array([[0.2857143], [0.8]]),
        )

    def it_knows_its_table_proportion_variances_to_help(self, request):
        weighted_counts = np.arange(12).reshape((2, 2, 3))
        np.testing.assert_almost_equal(
            _MrXCatMatrix(None, weighted_counts, None)._table_proportion_variances,
            np.array([[0.0, 0.0622222, 0.1155556], [0.1038062, 0.118416, 0.1322568]]),
        )


class Describe_MrXMrMatrix(object):
    """Unit test suite for `cr.cube.matrix._MrXMrMatrix` object."""

    def it_knows_its_columns_margin(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [[0, 8], [2, 7], [1, 7]],
                    [[2, 6], [6, 8], [3, 5]],
                ],
                [  # -- row 1 ---------------
                    [[4, 4], [1, 7], [8, 3]],
                    [[6, 2], [3, 5], [5, 2]],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, weighted_counts, None, None).columns_margin,
            np.array([[2, 8, 4], [10, 4, 13]]),
        )

    def it_knows_its_columns_pruning_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [[0, 8], [2, 7], [1, 7]],
                    [[2, 6], [6, 8], [3, 5]],
                ],
                [  # -- row 1 ---------------
                    [[4, 4], [1, 7], [8, 3]],
                    [[6, 2], [3, 5], [5, 2]],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, None, unweighted_counts, None).columns_pruning_base,
            np.array([12, 12, 17]),
        )

    def it_knows_its_rows_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [[0, 8], [2, 7], [1, 7]],
                    [[2, 6], [6, 8], [3, 5]],
                ],
                [  # -- row 1 ---------------
                    [[4, 4], [1, 7], [8, 3]],
                    [[6, 2], [3, 5], [5, 2]],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, None, unweighted_counts, None).rows_base,
            np.array([[8, 9, 8], [8, 8, 11]]),
        )

    def it_knows_its_rows_margin(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [[0, 8], [2, 7], [1, 7]],
                    [[2, 6], [6, 8], [3, 5]],
                ],
                [  # -- row 1 ---------------
                    [[4, 4], [1, 7], [8, 3]],
                    [[6, 2], [3, 5], [5, 2]],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, weighted_counts, None, None).rows_margin,
            np.array([[8, 9, 8], [8, 8, 11]]),
        )

    def it_knows_its_rows_pruning_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 ---------------
                    [[0, 8], [2, 7], [1, 7]],
                    [[2, 6], [6, 8], [3, 5]],
                ],
                [  # -- row 1 ---------------
                    [[4, 4], [1, 7], [8, 3]],
                    [[6, 2], [3, 5], [5, 2]],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, None, unweighted_counts, None).rows_pruning_base,
            np.array([25, 27]),
        )

    def it_knows_its_table_base(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 -------------
                    [  # -- selected ------
                        [0, 5],  # -- col 0
                        [1, 4],  # -- col 1
                    ],
                    [  # -- not selected --
                        [2, 3],  # -- col 0
                        [3, 1],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [4, 1],  # -- col 0
                        [5, 0],  # -- col 1
                    ],
                    [  # -- not selected --
                        [6, 9],  # -- col 0
                        [7, 6],  # -- col 1
                    ],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, None, unweighted_counts, None).table_base,
            np.array([[10, 9], [20, 18]]),
        )

    def it_knows_its_table_margin(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 -------------
                    [  # -- selected ------
                        [0, 5],  # -- col 0
                        [1, 4],  # -- col 1
                    ],
                    [  # -- not selected --
                        [2, 3],  # -- col 0
                        [3, 1],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [4, 1],  # -- col 0
                        [5, 0],  # -- col 1
                    ],
                    [  # -- not selected --
                        [6, 9],  # -- col 0
                        [7, 6],  # -- col 1
                    ],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, weighted_counts, None, None).table_margin,
            np.array([[10, 9], [20, 18]]),
        )

    def it_knows_its_unweighted_counts(self):
        unweighted_counts = np.array(
            [
                [  # -- row 0 -------------
                    [  # -- selected ------
                        [0, 8],  # -- col 0
                        [1, 7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [2, 6],  # -- col 0
                        [3, 5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [4, 4],  # -- col 0
                        [5, 3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [6, 2],  # -- col 0
                        [7, 1],  # -- col 1
                    ],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, None, unweighted_counts, None).unweighted_counts,
            np.array([[0, 1], [4, 5]]),
        )

    def it_knows_its_weighted_counts(self):
        weighted_counts = np.array(
            [
                [  # -- row 0 -------------
                    [  # -- selected ------
                        [0, 8],  # -- col 0
                        [1, 7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [2, 6],  # -- col 0
                        [3, 5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [4, 4],  # -- col 0
                        [5, 3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [6, 2],  # -- col 0
                        [7, 1],  # -- col 1
                    ],
                ],
            ]
        )
        np.testing.assert_equal(
            _MrXMrMatrix(None, weighted_counts, None, None).weighted_counts,
            np.array([[0, 1], [4, 5]]),
        )

    def it_knows_its_baseline_to_help(self, request):
        property_mock(
            request, _MrXMrMatrix, "_valid_row_idxs", return_value=np.array([0, 1])
        )
        counts_with_missings = np.array(
            [
                [[[0, 8], [1, 7]], [[2, 6], [3, 5]]],
                [[[4, 4], [5, 3]], [[6, 2], [7, 1]]],
            ]
        )

        np.testing.assert_almost_equal(
            _MrXMrMatrix(None, None, None, counts_with_missings)._baseline,
            np.array([[0.5, 0.5], [0.5, 0.5]]),
        )

    def it_knows_its_table_proportion_variances_to_help(self):
        weighted_counts = np.arange(24).reshape((2, 2, 3, 2))
        np.testing.assert_almost_equal(
            _MrXMrMatrix(None, weighted_counts, None)._table_proportion_variances,
            np.array([[0.0, 0.0826446, 0.1155556], [0.1560874, 0.16, 0.1630506]]),
        )



# encoding: utf-8

"""Integration-test suite for `cr.cube.cube` module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pytest

from cr.cube.cube import (
    Cube,
    _MeanMeasure,
    _Measures,
    _UnweightedCountMeasure,
    _WeightedCountMeasure,
)
from cr.cube.dimension import _ApparentDimensions, AllDimensions
from cr.cube.enums import DIMENSION_TYPE as DT

from ..fixtures import (
    CR,  # cube-response
    OL,  # overlaps
    NA,  # numeric-array
)


class DescribeIntegratedCube(object):
    """Integration-test suite for `cr.cube.cube.Cube` object."""

    def it_provides_values_for_cat_x_cat(self):
        cube = Cube(CR.CAT_X_CAT)

        assert cube.__repr__() == "Cube(name='v4', dimension_types='CAT x CAT')"
        assert cube.counts == pytest.approx(np.array([[5, 2], [5, 3]]))
        assert cube.counts_with_missings == pytest.approx(
            np.array([[5, 3, 2, 0], [5, 2, 3, 0], [0, 0, 0, 0]])
        )
        assert cube.cube_index == 0
        assert cube.description == "Pet Owners"
        assert cube.dimension_types == (DT.CAT, DT.CAT)
        assert isinstance(cube.dimensions, _ApparentDimensions)
        assert cube.has_weighted_counts is False
        assert cube.missing == 5
        assert cube.name == "v4"
        assert cube.ndim == 2
        assert cube.population_fraction == 1.0
        assert cube.title == "Pony Owners"
        assert cube.unweighted_counts == pytest.approx(np.array([[5, 2], [5, 3]]))
        assert cube.weighted_counts is None

    @pytest.mark.parametrize(
        "cube_response, expected_dim_types",
        (
            (CR.CA_SUBVAR_HS_X_MR_X_CA_CAT, (DT.CA_SUBVAR, DT.MR, DT.CA_CAT)),
            (CR.CAT_X_LOGICAL, (DT.CAT, DT.LOGICAL)),
            (CR.LOGICAL_UNIVARIATE, (DT.LOGICAL,)),
        ),
    )
    def it_provides_access_to_its_dimensions(self, cube_response, expected_dim_types):
        cube = Cube(cube_response)

        dimension_types = tuple(d.dimension_type for d in cube.dimensions)

        assert dimension_types == expected_dim_types

    def it_provides_array_for_single_valid_cat_CAT_X_MR(self):
        """No pruning needs to happen, because pruning is based on unweighted counts:
        >>> cube.unweighted_counts
        array([[[0, 108],
                [14, 94],
                [94, 14]]])

        so even though the weighted counts are all zeroes:

        >>> cube.counts
        array([[[0, 0],
                [0, 0],
                [0, 0]]])

        we expect [[0, 0, 0]] as the result; no zero gets pruned because no unweighted
        count is zero.
        """
        transforms = {
            "rows_dimension": {"prune": True},
            "columns_dimension": {"prune": True},
        }
        slice_ = Cube(CR.CAT_X_MR_SENTRY, transforms=transforms).partitions[0]
        np.testing.assert_array_equal(slice_.counts, np.array([[0, 0, 0]]))

    def it_provides_pruned_array_for_CA_CAT_x_CA_SUBVAR(self):
        transforms = {
            "rows_dimension": {"prune": True},
            "columns_dimension": {"prune": True},
        }
        slice_ = Cube(CR.CA_CAT_X_CA_SUBVAR, transforms=transforms).partitions[0]
        np.testing.assert_array_equal(
            slice_.column_proportions,
            np.array(
                [
                    [0.19012797074954296, 0.10494203782794387],
                    [0.2528945764777575, 0.2190359975594875],
                    [0.16514320536258378, 0.1720561317876754],
                    [0.29859841560024375, 0.4069554606467358],
                    [0.09323583180987204, 0.09701037217815742],
                ]
            ),
        )

    def it_provides_valid_counts_for_NUM_ARRAY_GROUPED_BY_CAT(self):
        cube = Cube(NA.NUM_ARR_MEANS_GROUPED_BY_CAT)

        assert cube.covariance is None
        assert cube.unweighted_valid_counts == pytest.approx(
            np.array([[3, 2], [3, 1], [1, 1]])
        )

    def it_provides_valid_counts_summary_for_NUM_ARRAY_GROUPED_BY_CAT(self):
        cube = Cube(NA.NUM_ARR_MEANS_GROUPED_BY_CAT)

        np.testing.assert_array_equal(cube.valid_counts_summary, [5, 4, 2])

    def and_it_returns_empty_array_for_summary_if_valid_counts_are_not_available(self):
        cube = Cube(CR.CAT_X_CAT)

        np.testing.assert_array_equal(cube.valid_counts_summary, [])

    def it_provides_n_responses_for_NUM_ARRAY_GROUPED_BY_CAT(self):
        cube = Cube(NA.NUM_ARR_MEANS_GROUPED_BY_CAT)

        assert cube.n_responses == 5

    def it_provides_multiple_measures_for_NUM_ARRAY_GROUPED_BY_CAT(self):
        cube = Cube(NA.NUM_ARR_MULTI_NUMERIC_MEASURES_GROUPED_BY_CAT)

        assert cube.sums.tolist() == [
            [183.0, 105.0, 43.0, 31.0],
            [230.0, 71.0, 79.0, 123.0],
            [25.0, 58.0, 148.0, 72.0],
        ]
        assert cube.means == pytest.approx(
            np.array(
                [
                    [61.0, 52.5, 14.333333, 10.333333],
                    [76.6666667, 35.5, 26.333333, 41.0],
                    [8.333333, 29.0, 49.333333, 24.0],
                ]
            )
        )
        assert cube.covariance == pytest.approx(
            np.array(
                [
                    [
                        [1623.0, 1107.5, 155.5],
                        [312.5, -362.5, 400.0],
                        [217.333333, 124.333333, 11.333333],
                        [25.333333, 73.0, 12.0],
                    ],
                    [
                        [1107.5, 758.333333, 124.166666],
                        [-362.5, 420.5, -464.0],
                        [124.333333, 134.333333, -189.666666],
                        [73.0, 937.0, -639.5],
                    ],
                    [
                        [155.5, 124.166666, 140.333333],
                        [400.0, -464.0, 512.0],
                        [11.333333, -189.666666, 609.333333],
                        [12.0, -639.5, 631.0],
                    ],
                ]
            )
        )

    def test_optional_covariance_cube_measure(self):
        cube = Cube(NA.NUM_ARR_MEANS_GROUPED_BY_CAT)

        assert cube.covariance is None


class DescribeIntegrated_Measures(object):
    """Integration-tests that exercise the `cr.cube.cube._Measures` object."""

    def it_provides_access_to_the_overlaps_measure(self):
        cube_dict = OL.CAT_X_MR_SUB_X_MR_SEL
        measures = _Measures(
            cube_dict,
            AllDimensions(dimension_dicts=cube_dict["result"]["dimensions"]),
        )

        overlaps = measures.overlaps

        assert type(overlaps).__name__ == "_OverlapMeasure"

    def but_only_when_the_cube_response_contains_overlaps(self):
        cube_dict = CR.CAT_X_CAT
        measures = _Measures(cube_dict, None)

        overlaps = measures.overlaps

        assert overlaps is None

    def it_provides_access_to_the_mean_measure(self):
        cube_dict = CR.CAT_X_CAT_MEAN_WGTD
        measures = _Measures(
            cube_dict,
            AllDimensions(dimension_dicts=cube_dict["result"]["dimensions"]),
        )

        means = measures.means

        assert type(means).__name__ == "_MeanMeasure"

    def but_only_when_the_cube_response_contains_means(self):
        cube_dict = CR.CAT_X_CAT
        measures = _Measures(cube_dict, None)

        means = measures.means

        assert means is None

    def it_provides_the_means_missing_count_when_means_are_available(self):
        cube_dict = CR.CAT_X_CAT_MEAN_WGTD
        measures = _Measures(
            cube_dict,
            AllDimensions(dimension_dicts=cube_dict["result"]["dimensions"]),
        )
        missing_count = measures.missing_count
        assert missing_count == 3

    def it_provides_the_means_missing_count_when_sum_are_available(self):
        cube_dict = CR.SUM_CAT_X_MR
        measures = _Measures(
            cube_dict,
            AllDimensions(dimension_dicts=cube_dict["result"]["dimensions"]),
        )
        missing_count = measures.missing_count
        assert missing_count == 1

    def but_provides_the_general_missing_count_otherwise(self):
        measures = _Measures(CR.CAT_X_CAT, None)
        missing_count = measures.missing_count
        assert missing_count == 5

    @pytest.mark.parametrize(
        "cube_dict, expected_value",
        (
            # ---filtered case---
            (CR.CAT_X_CAT_FILT, 0.254),
            # ---unfiltered case---
            (CR.CAT_X_CAT, 1.0),
            # ---complete cases---
            (CR.CAT_X_CAT_FILT_COMPLETE, 0.5760869565217391),
        ),
    )
    def it_knows_the_population_fraction(self, cube_dict, expected_value):
        measures = _Measures(cube_dict, None)

        population_fraction = measures.population_fraction

        assert population_fraction == expected_value

    def it_provides_access_to_the_unweighted_count_measure(self):
        measures = _Measures(None, None)

        unweighted_counts = measures.unweighted_counts

        assert type(unweighted_counts).__name__ == "_UnweightedCountMeasure"

    @pytest.mark.parametrize(
        "cube_dict, expected_type_name",
        (
            # ---weighted case---
            (CR.CAT_X_CAT_WGTD, "_WeightedCountMeasure"),
            # ---unweighted case---
            (CR.CAT_X_CAT, "NoneType"),
        ),
    )
    def it_provides_access_to_wgtd_count_measure(self, cube_dict, expected_type_name):
        measures = _Measures(
            cube_dict,
            AllDimensions(dimension_dicts=cube_dict["result"]["dimensions"]),
        )

        weighted_counts = measures.weighted_counts

        assert type(weighted_counts).__name__ == expected_type_name


class DescribeIntegrated_MeanMeasure(object):
    def it_provides_access_to_its_raw_cube_array(self):
        cube_dict = CR.CAT_X_CAT_MEAN_WGTD
        cube = Cube(cube_dict)
        measure = _MeanMeasure(cube_dict, cube._all_dimensions)

        raw_cube_array = measure.raw_cube_array

        np.testing.assert_array_almost_equal(
            raw_cube_array,
            [
                [52.78205128, 49.90697674, np.nan, np.nan, np.nan],
                [50.43654822, 48.20100503, np.nan, np.nan, np.nan],
                [51.56435644, 47.60283688, np.nan, np.nan, np.nan],
                [58.0, 29.0, np.nan, np.nan, np.nan],
                [37.53846154, 39.45238095, np.nan, np.nan, np.nan],
                [36.66666667, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
        )

    def it_handles_cat_x_mr_with_means(self):
        slice_ = Cube(CR.MEANS_CAT_X_MR).partitions[0]
        assert slice_.column_labels.tolist() == [
            "Denmark",
            "Finland",
            "Iceland",
            "Norway",
            "Sweden",
        ]

    def it_handles_means_cat_hs_x_cat_hs(self):
        slice_ = Cube(CR.MEANS_CAT_HS_X_CAT_HS).partitions[0]

        means = slice_.means

        np.testing.assert_array_almost_equal(
            means,
            [
                [41.96875, 30.875, 25.66666667, np.nan, 42.0],
                [51.51515152, 47.95555556, 45.44444444, np.nan, 45.0952381],
                [46.17088608, 44.55504587, 48.09090909, np.nan, 50.8],
                [np.nan, np.nan, np.nan, np.nan, np.nan],
                [44.03030303, 45.21568627, 54.53333333, np.nan, 56.19512195],
                [45.64516129, 47.41428571, 46.89361702, np.nan, 55.27894737],
                [34.20408163, 43.2745098, 41.2, np.nan, 35.26086957],
            ],
        )


class DescribeIntegrated_UnweightedCountMeasure(object):
    def it_provides_access_to_its_raw_cube_array(self):
        cube_dict = CR.CAT_X_CAT
        cube = Cube(cube_dict)
        measure = _UnweightedCountMeasure(cube_dict, cube._all_dimensions)

        raw_cube_array = measure.raw_cube_array

        np.testing.assert_array_almost_equal(
            raw_cube_array, [[5, 3, 2, 0], [5, 2, 3, 0], [0, 0, 0, 0]]
        )


class DescribeIntegrated_WeightedCountMeasure(object):
    def it_provides_access_to_its_raw_cube_array(self):
        cube_dict = CR.CAT_X_CAT_WGTD
        cube = Cube(cube_dict)
        measure = _WeightedCountMeasure(cube_dict, cube._all_dimensions)

        raw_cube_array = measure.raw_cube_array

        np.testing.assert_array_almost_equal(
            raw_cube_array,
            [
                [32.9, 87.6, 176.2, 117.5, 72.1, 13.4, 0.0, 0.0, 0.0],
                [38.8, 94.1, 199.0128, 102.9, 38.8305, 26.2135, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
        )



# encoding: utf-8

"""The `Assembler` object provides the external interface for this module.

Its name derives from its role to "assemble" a finished 2D array ("matrix") for a
particular measure from the base measure values and inserted subtotals, to reorder the
rows and columns according to the dimension *order* transforms, and to hide rows and
columns that are either hidden by the user or "pruned" because they contain no
observations.
"""

from __future__ import division

import numpy as np
from scipy.stats import norm

from cr.cube.collator import (
    ExplicitOrderCollator,
    PayloadOrderCollator,
    SortByValueCollator,
)
from cr.cube.enums import (
    COLLATION_METHOD as CM,
    DIMENSION_TYPE as DT,
    MARGINAL,
    MARGINAL_ORIENTATION as MO,
    MEASURE as M,
)
from cr.cube.matrix.cubemeasure import BaseCubeResultMatrix
from cr.cube.matrix.measure import SecondOrderMeasures
from cr.cube.matrix.subtotals import (
    NanSubtotals,
    SumSubtotals,
    TableStdErrSubtotals,
)
from cr.cube.util import lazyproperty


class Assembler(object):
    """Provides measure and margin methods for a cube-slice.

    An assembled matrix is a 2D ndarray reflecting all ordering, insertion, and hiding
    transforms applied to the dimensions. An assembled margin is often a 1D ndarray
    which is similarly formed from inserted values, ordered, and value hiding applied.

    `cube` is the `cr.cube.Cube` object containing the data for this matrix. Note that
    not all the data in `cube` will necessarily be used by this matrix. When `cube` is
    more than 2-dimensional, it is "sliced" and each slice gets its own matrix (and
    `_Slice` object).

    `dimensions` is a pair (2-tuple) of (rows_dimension, columns_dimension) Dimension
    objects. These are always the last two dimensions of `cube` but may and often do
    have transformations applied that are not present on the `cube` dimensions from
    which they derive.

    `slice_idx` is an int offset indicating which portion of `cube` data to use for this
    matrix. There is one slice for each element of the first cube dimension (the "table"
    dimension) when the cube has more than two dimensions.
    """

    def __init__(self, cube, dimensions, slice_idx):
        self._cube = cube
        self._dimensions = dimensions
        self._slice_idx = slice_idx

    @lazyproperty
    def column_comparable_counts(self):
        """2D np.float64 ndarray of counts comparable in the column dimension.

        Comparable in the "column" dimension means that it's not a column subtotal
        difference because these are not considered comparable since their bases
        aren't the same. The values along these subtotal differences are defined as
        np.nan.
        """
        return self._assemble_matrix(self._measures.column_comparable_counts.blocks)

    @lazyproperty
    def column_index(self):
        """2D np.float64 ndarray of column-index "percentage" for each table cell."""
        return self._assemble_matrix(
            NanSubtotals.blocks(self._cube_result_matrix.column_index, self._dimensions)
        )

    @lazyproperty
    def column_labels(self):
        """1D str ndarray of column name for each matrix column.

        These are suitable for use as column headings; labels for subtotal columns
        appear in the sequence and labels are ordered to correspond with their
        respective data column.
        """
        return self._dimension_labels(self._columns_dimension, self._column_order)

    @lazyproperty
    def column_proportions(self):
        """2D np.float64 ndarray of column-proportion for each matrix cell.

        This is the proportion of the weighted-count for cell to the weighted-N of the
        column the cell appears in (aka. column-margin). Generally a number between 0.0
        and 1.0 inclusive, but subtotal differences can be between -1.0 and 1.0
        inclusive.
        """
        return self._assemble_matrix(self._measures.column_proportions.blocks)

    @lazyproperty
    def column_share_sum(self):
        """2D optional np.float64 ndarray of column share of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum cube-measure.
        """
        return self._assemble_matrix(self._measures.column_share_sum.blocks)

    @lazyproperty
    def column_unweighted_bases(self):
        """2D np.float64 ndarray of unweighted col-proportions denominator per cell."""
        return self._assemble_matrix(self._measures.column_unweighted_bases.blocks)

    @lazyproperty
    def column_weighted_bases(self):
        """2D np.float64 ndarray of column-proportions denominator for each cell."""
        return self._assemble_matrix(self._measures.column_weighted_bases.blocks)

    @lazyproperty
    def columns_base(self):
        """1D/2D np.float64 ndarray of unweighted-N for each slice column/cell."""
        # --- an MR_X slice produces a 2D column-base (each cell has its own N) ---
        rows_dim_type = self._rows_dimension.dimension_type
        columns_base = self._measures.columns_base
        if rows_dim_type in (DT.MR_SUBVAR, DT.NUM_ARRAY):
            return self._assemble_matrix(
                SumSubtotals.blocks(columns_base, self._dimensions, diff_cols_nan=True)
            )
        # --- otherwise columns-base is a vector ---
        return self._assemble_vector(
            columns_base, self._column_subtotals, self._column_order, diffs_nan=True
        )

    @lazyproperty
    def columns_dimension_numeric_values(self):
        """1D optional np.int/float64 ndarray of numeric-value for each column element.

        A value of np.nan appears for a column element without a numeric-value. All
        subtotal columns have a value of np.nan (subtotals have no numeric value).
        """
        elements = self._columns_dimension.valid_elements
        return np.array(
            [
                (elements[idx].numeric_value if idx >= 0 else np.nan)
                for idx in self._column_order
            ]
        )

    @lazyproperty
    def columns_margin(self):
        """1D/2D np.float64 ndarray of weighted-N for each column of this slice."""
        # --- an MR_X slice produces a 2D columns-margin (each cell has its own N) ---
        if self._rows_dimension.dimension_type == DT.MR_SUBVAR:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self._cube_result_matrix.columns_margin,
                    self._dimensions,
                    diff_cols_nan=True,
                )
            )

        # --- otherwise columns-base is a vector ---
        return self._assemble_vector(
            self._cube_result_matrix.columns_margin,
            self._column_subtotals,
            self._column_order,
            diffs_nan=True,
        )

    @lazyproperty
    def columns_margin_proportion(self):
        """1D/2D np.float64 ndarray of weighted-prop for each column of this slice."""
        # --- an MR_X slice produces a 2D columns-margin (each cell has its own N) ---
        denominator = self._cube_result_matrix.table_margin
        if self._rows_dimension.dimension_type == DT.MR_SUBVAR:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self._cube_result_matrix.columns_margin / denominator,
                    self._dimensions,
                )
            )

        # --- otherwise columns-base is a vector ---
        return self._assemble_vector(
            self._cube_result_matrix.columns_margin / denominator,
            self._column_subtotals,
            self._column_order,
        )

    @lazyproperty
    def columns_scale_mean(self):
        """Optional 1D column scale mean marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_mean)

    @lazyproperty
    def columns_scale_mean_stddev(self):
        """Optional 1D column scale mean std dev marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_mean_stddev)

    @lazyproperty
    def columns_scale_median(self):
        """Optional 1D column scale median marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_median)

    @lazyproperty
    def inserted_column_idxs(self):
        """tuple of int index of each subtotal column in slice."""
        # --- insertions have a negative idx in their order sequence ---
        return tuple(i for i, col_idx in enumerate(self._column_order) if col_idx < 0)

    @lazyproperty
    def inserted_row_idxs(self):
        """tuple of int index of each subtotal row in slice."""
        # --- insertions have a negative idx in their order sequence ---
        return tuple(i for i, row_idx in enumerate(self._row_order) if row_idx < 0)

    @lazyproperty
    def means(self):
        """2D optional np.float64 ndarray of mean for each cell.

        Raises `ValueError` if the cube-result does not include a means cube-measure.
        """
        return self._assemble_matrix(self._measures.means.blocks)

    def pairwise_indices(self, alpha, only_larger):
        """2D optional ndarray of tuple of int column-idxs pairwise-t threshold.

        Raises `ValueError if the cube-result does not include `overlaps` cube-measures.
        """
        return self._assemble_matrix(
            self._measures.pairwise_indices(alpha, only_larger).blocks
        )

    def pairwise_means_indices(self, alpha, only_larger):
        """2D optional ndarray of tuple of int column-idxs means pairwise-t threshold.

        Raises `ValueError if the cube-result does not include `means` cube-measures.
        """
        assembled_matrix = self._assemble_matrix(
            self._measures.pairwise_means_indices(alpha, only_larger).blocks
        )
        ord_map = {k: v for v, k in enumerate(self._column_order)}

        return np.array(
            [
                [
                    tuple(ord_map[idx] for idx in idxs) if idxs is not None else None
                    for idxs in row
                ]
                for row in assembled_matrix
            ],
            dtype=object,
        )

    def pairwise_significance_p_vals(self, subvar_idx):
        """2D optional np.float64 ndarray of overlaps-p_vals matrices for subvar idx.

        Raises `ValueError if the cube-result does not include `overlaps`
        and `valid_overlaps` cube-measures.
        """
        return self._assemble_matrix(
            self._measures.pairwise_p_vals_for_subvar(subvar_idx).blocks
        )

    def pairwise_significance_t_stats(self, subvar_idx):
        """2D optional np.float64 ndarray of overlaps-t_stats matrices for subvar idx.

        Raises `ValueError if the cube-result does not include `overlaps`
        and `valid_overlaps` cube-measures.
        """
        return self._assemble_matrix(
            self._measures.pairwise_t_stats_for_subvar(subvar_idx).blocks
        )

    def pairwise_significance_means_p_vals(self, column_idx):
        """2D optional np.float64 ndarray of mean difference p_vals for column idx.

        Raises `ValueError if the cube-result does not include `mean` cube-measures.
        """
        base_column_idx = self._column_order[column_idx]
        return self._assemble_matrix(
            self._measures.pairwise_significance_means_p_vals(base_column_idx).blocks
        )

    def pairwise_significance_means_t_stats(self, column_idx):
        """2D optional np.float64 ndarray of mean difference t_stats for column idx.

        Raises `ValueError if the cube-result does not include `mean` cube-measures.
        """
        base_column_idx = self._column_order[column_idx]
        return self._assemble_matrix(
            self._measures.pairwise_significance_means_t_stats(base_column_idx).blocks
        )

    @lazyproperty
    def pvalues(self):
        """2D np.float64/np.nan ndarray of p-value for each matrix cell."""
        return 2 * (1 - norm.cdf(np.abs(self.zscores)))

    @lazyproperty
    def row_comparable_counts(self):
        """2D np.float64 ndarray of counts comparable in the row dimension.

        Comparable in the "row" dimension means that it's not a row subtotal
        difference because these are not considered comparable since their bases
        aren't the same. The values along these subtotal differences are defined as
        np.nan.
        """
        return self._assemble_matrix(self._measures.row_comparable_counts.blocks)

    @lazyproperty
    def row_labels(self):
        """1D str ndarray of row name for each matrix row.

        These are suitable for use as row headings; labels for subtotal rows appear in
        the sequence and labels are ordered to correspond with their respective data
        row.
        """
        return self._dimension_labels(self._rows_dimension, self._row_order)

    @lazyproperty
    def row_proportions(self):
        """2D np.float64 ndarray of row-proportion for each matrix cell.

        This is the proportion of the weighted-count for cell to the weighted-N of the
        row the cell appears in (aka. row-margin). Always a number between 0.0 and
        1.0 inclusive, but subtotal differences can be between -1.0 and 1.0 inclusive.
        """
        return self._assemble_matrix(self._measures.row_proportions.blocks)

    @lazyproperty
    def row_share_sum(self):
        """2D optional np.float64 ndarray of row share of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum cube-measure.
        """
        return self._assemble_matrix(self._measures.row_share_sum.blocks)

    @lazyproperty
    def row_unweighted_bases(self):
        """2D np.float64 ndarray of unweighted row-proportions denominator per cell."""
        return self._assemble_matrix(self._measures.row_unweighted_bases.blocks)

    @lazyproperty
    def row_weighted_bases(self):
        """2D np.float64 ndarray of row-proportions denominator for each cell."""
        return self._assemble_matrix(self._measures.row_weighted_bases.blocks)

    @lazyproperty
    def rows_base(self):
        """1D/2D np.float64 ndarray of unweighted-N for each slice row/cell."""
        # --- an X_MR slice produces a 2D row-base (each cell has its own N) ---
        if self._columns_dimension.dimension_type == DT.MR_SUBVAR:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self._cube_result_matrix.rows_base,
                    self._dimensions,
                    diff_rows_nan=True,
                )
            )

        # --- otherwise rows-base is a vector ---
        return self._assemble_vector(
            self._cube_result_matrix.rows_base,
            self._row_subtotals,
            self._row_order,
            diffs_nan=True,
        )

    @lazyproperty
    def rows_dimension_fills(self):
        """tuple of RGB str like "#def032" fill color for each row in slice."""
        elements = self._rows_dimension.valid_elements
        return tuple(
            (elements[idx].fill if idx >= 0 else None) for idx in self._row_order
        )

    @lazyproperty
    def rows_dimension_numeric_values(self):
        """1D optional np.int/float64 ndarray of numeric-value for each row element.

        A value of np.nan appears for a row element without a numeric-value. All
        subtotal rows have a value of np.nan (subtotals have no numeric value).
        """
        elements = self._rows_dimension.valid_elements
        return np.array(
            [
                (elements[idx].numeric_value if idx >= 0 else np.nan)
                for idx in self._row_order
            ]
        )

    @lazyproperty
    def rows_margin(self):
        """1D/2D np.float64 ndarray of weighted-N for each column of this slice.

        This value is a 2D array if one or both of the slice dimensions are MR. This is
        because each cell has a distinct margin in such a slice.
        """
        # --- an X_MR slice produces a 2D rows-margin (each cell has its own N) ---
        if self._columns_dimension.dimension_type == DT.MR_SUBVAR:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self._cube_result_matrix.rows_margin,
                    self._dimensions,
                    diff_rows_nan=True,
                )
            )

        # --- otherwise rows-margin is a vector ---
        return self._assemble_vector(
            self._cube_result_matrix.rows_margin,
            self._row_subtotals,
            self._row_order,
            diffs_nan=True,
        )

    @lazyproperty
    def rows_margin_proportion(self):
        """1D/2D np.float64 ndarray of weighted-proportion for each slice row/cell."""
        # --- an X_MR slice produces a 2D rows-margin (each cell has its own N) ---
        denominator = self._cube_result_matrix.table_margin
        if self._columns_dimension.dimension_type == DT.MR_SUBVAR:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self._cube_result_matrix.rows_margin / denominator,
                    self._dimensions,
                )
            )

        # --- otherwise rows-margin is a vector ---
        return self._assemble_vector(
            self._cube_result_matrix.rows_margin / denominator,
            self._row_subtotals,
            self._row_order,
        )

    @lazyproperty
    def rows_scale_mean(self):
        """Optional 1D Row scale median marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_mean)

    @lazyproperty
    def rows_scale_mean_stddev(self):
        """Optional 1D row scale mean std dev marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_mean_stddev)

    @lazyproperty
    def rows_scale_median(self):
        """Optional 1D Row scale median marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_median)

    @lazyproperty
    def sums(self):
        """2D optional np.float64 ndarray of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum cube-measure.
        """
        return self._assemble_matrix(self._measures.sums.blocks)

    @lazyproperty
    def stddev(self):
        """2D optional np.float64 ndarray of stddev for each cell.

        Raises `ValueError` if the cube-result does not include a stddev cube-measure.
        """
        return self._assemble_matrix(self._measures.stddev.blocks)

    @lazyproperty
    def table_base(self):
        """Scalar, 1D, or 2D ndarray of np.float64 unweighted-N for this slice.

        This value has four distinct forms, depending on the slice dimensions:

            * MR_X_MR - 2D ndarray with a distinct table-base value per cell.
            * MR_X - 1D ndarray of value per *row* when only rows dimension is MR.
            * X_MR - 1D ndarray of value per *column* when only columns dimension is MR.
            * CAT_X_CAT - scalar float value when slice has no MR dimension.

        """
        # --- an MR dimension cannot have subtotals, and a non-scalar table base only
        # --- arises in an MR dimension case, so a non-scalar table-base never has
        # --- subtotals and we need only operate on the base values. There is no need to
        # --- assemble with subtotals, however the values of a non-scalar table-base do
        # --- still need to be ordered (which includes hiding).
        base_table_base = self._cube_result_matrix.table_base

        # --- MR_X_MR slice produces a 2D table-base (each cell has its own N) ---
        if (
            self._rows_dimension.dimension_type == DT.MR_SUBVAR
            and self._columns_dimension.dimension_type == DT.MR_SUBVAR
        ):
            return base_table_base[np.ix_(self._row_order, self._column_order)]

        # --- CAT_X_MR slice produces a 1D table-base (each column has its own N) ---
        if self._columns_dimension.dimension_type == DT.MR_SUBVAR:
            return base_table_base[self._column_order]

        # --- MR_X_CAT slice produces a 1D table-base (each row has its own N) ---
        if self._rows_dimension.dimension_type == DT.MR_SUBVAR:
            return base_table_base[self._row_order]

        # --- CAT_X_CAT produces scalar table-base ---
        return self._cube_result_matrix.table_base

    @lazyproperty
    def table_base_unpruned(self):
        """np.float64 scalar or a 1D or 2D ndarray of np.float64 representing table base.

        This value includes hidden vectors, those with either a hide transform on
        their element or that have been pruned (because their base (N) is zero). Also,
        it does *not* include inserted subtotals. This does not affect a scalar value
        but when the return value is an ndarray, the shape may be different than the
        array returned by `.table_base`.

        A multiple-response (MR) dimension produces an array of table-base values
        because each element (subvariable) of the dimension represents a logically
        distinct question which may not have been asked of all respondents. When both
        dimensions are MR, the return value is a 2D ndarray. A CAT_X_CAT matrix produces
        a scalar value for this property.
        """
        # TODO: This name is misleading. It's not only "unpruned" it's "before_hiding"
        # (of either kind, prune or hide-transform). But the real problem is having this
        # interface property at all. The need for this is related to expressing ranges
        # for base and margin in cubes that have an MR dimension. The real solution is
        # to compute ranges in `cr.cube` rather than leaking this sort of internal
        # detail through the interface and making the client compute those for
        # themselves. So this will require reconstructing that "show-ranges" requirement
        # and either adding some sort of a `.range` property that returns a sequence of
        # (min, max) tuples, or maybe just returning margin or base as tuples when
        # appropriate and having something like a `.margin_is_ranges` predicate the
        # client can switch on to control their rendering.
        return self._cube_result_matrix.table_base

    @lazyproperty
    def table_margin(self):
        """Scalar, 1D, or 2D ndarray of np.float64 weighted-N for this slice.

        This value has four distinct forms, depending on the slice dimensions:

            * MR_X_MR - 2D ndarray with a distinct table-margin value per cell.
            * MR_X - 1D ndarray of value per *row* when only rows dimension is MR.
            * X_MR - 1D ndarray of value per *column* when only columns dimension is MR.
            * CAT_X_CAT - scalar float value when slice has no MR dimension.

        """
        # NOTE: This all works out even though it doesn't include subtotals and isn't
        # assembled, because an MR dimension can't have subtotals. Because the
        # dimensionality of table-margin "follows" the MR-dimension(s), it follows that
        # a table-margin vector or matrix never has subtotals.

        base_table_margin = self._cube_result_matrix.table_margin

        # --- MR_X_MR slice produces a 2D table-margin (each cell has its own N) ---
        if (
            self._rows_dimension.dimension_type == DT.MR_SUBVAR
            and self._columns_dimension.dimension_type == DT.MR_SUBVAR
        ):
            return base_table_margin[np.ix_(self._row_order, self._column_order)]

        # --- MR_X_CAT slice produces a 1D table-margin (each row has its own N) ---
        if self._rows_dimension.dimension_type == DT.MR_SUBVAR:
            return base_table_margin[self._row_order]

        # --- CAT_X_MR slice produces a 1D table-margin (each column has its own N) ---
        if self._columns_dimension.dimension_type == DT.MR_SUBVAR:
            return base_table_margin[self._column_order]

        # --- CAT_X_CAT produces scalar table-margin ---
        return base_table_margin

    @lazyproperty
    def table_margin_unpruned(self):
        """np.float64 scalar or a 1D or 2D ndarray of np.float64 table margin.

        This value includes hidden vectors, those with either a hide transform on
        their element or that have been pruned (because their base (N) is zero). Also,
        it does not include inserted subtotals. This
        does not affect a scalar value but when the return value is an ndarray, the
        shape may be different than the array returned by `.table_margin`.

        A matrix with a multiple-response (MR) dimension produces an array of
        table-margin values because each element (subvariable) of the dimension
        represents a logically distinct question which may not have been asked of all
        respondents. When both dimensions are MR, the return value is a 2D ndarray.
        A CAT_X_CAT matrix produces a scalar value for this property.
        """
        # TODO: see TODO in `.table_base_unpruned`
        return self._cube_result_matrix.table_margin

    @lazyproperty
    def table_stderrs(self):
        """2D np.float64 ndarray of std-error of table-percent for each matrix cell."""
        return self._assemble_matrix(
            TableStdErrSubtotals.blocks(
                self._cube_result_matrix.table_stderrs,
                self._dimensions,
                self._cube_result_matrix,
            )
        )

    @lazyproperty
    def table_unweighted_bases(self):
        """2D np.float64 ndarray of unweighted table-proportion denominator per cell."""
        return self._assemble_matrix(self._measures.table_unweighted_bases.blocks)

    @lazyproperty
    def table_weighted_bases(self):
        """2D np.float64 ndarray of weighted table-proportion denominator per cell."""
        return self._assemble_matrix(self._measures.table_weighted_bases.blocks)

    @lazyproperty
    def total_share_sum(self):
        """2D optional np.float64 ndarray of total share of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum
        cube-measure.
        """
        return self._assemble_matrix(self._measures.total_share_sum.blocks)

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each cell."""
        return self._assemble_matrix(self._measures.unweighted_counts.blocks)

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each cell."""
        return self._assemble_matrix(self._measures.weighted_counts.blocks)

    @lazyproperty
    def zscores(self):
        """2D np.float64 ndarray of std-res value for each cell of matrix.

        A z-score is also known as a *standard score* and is the number of standard
        deviations above (positive) or below (negative) the population mean a cell's
        value is.
        """
        return self._assemble_matrix(self._measures.zscores.blocks)

    def _assemble_marginal(self, marginal):
        """Optional 1D ndarray created from a marginal.

        The assembled marginal is the shape of either a row or column (determined by
        `marginal.orientation`), and with the ordering that's applied to those
        dimensions.

        It is None when the marginal is not defined (`marginal._is_defined`).
        """
        if not marginal.is_defined:
            return None

        order = (
            self._row_order if marginal.orientation == MO.ROWS else self._column_order
        )

        return np.hstack(marginal.blocks)[order]

    def _assemble_matrix(self, blocks):
        """Return 2D ndarray matrix assembled from `blocks`.

        The assembled matrix includes inserted vectors (rows and columns), has hidden
        vectors removed, and is ordered by whatever sort method is applied in the
        dimension transforms.
        """
        # --- These are assembled into a single 2D array, and then rearranged based on
        # --- row and column orders. All insertion, ordering, and hiding transforms are
        # --- reflected in the row and column orders. They each include (negative)
        # --- insertion idxs, hidden and pruned vector indices have been removed, and
        # --- the ordering method has been applied to determine the sequence each idx
        # --- appears in. This directly produces a final array that is exactly the
        # --- desired output.
        return np.block(blocks)[np.ix_(self._row_order, self._column_order)]

    def _assemble_vector(self, base_vector, subtotals, order, diffs_nan=False):
        """Return 1D ndarray of `base_vector` with inserted `subtotals`, in `order`.

        Each subtotal value is the result of applying np.sum to the addends and
        subtrahends extracted from `base_vector` according to the `addend_idxs`
        and `subtrahend_idxs` property of each subtotal in `subtotals`. The returned
        array is arranged by `order`, including possibly removing hidden or pruned
        values.
        """
        # TODO: This works for "sum" and "diff" subtotals, because either we set to
        # nan or add & subtract, but a fuller solution will probably get the subtotal
        # values from a _BaseSubtotals subclass.
        vector_subtotals = np.array(
            [
                np.nan
                if diffs_nan and len(subtotal.subtrahend_idxs) > 0
                else np.sum(base_vector[subtotal.addend_idxs])
                - np.sum(base_vector[subtotal.subtrahend_idxs])
                for subtotal in subtotals
            ]
        )
        return np.hstack([base_vector, vector_subtotals])[order]

    @lazyproperty
    def _column_order(self):
        """1D np.int64 ndarray of signed int idx for each assembled column.

        Negative values represent inserted subtotal-column locations.
        """
        return _BaseOrderHelper.column_display_order(self._dimensions, self._measures)

    @lazyproperty
    def _column_subtotals(self):
        """Sequence of _Subtotal object for each inserted column."""
        return self._columns_dimension.subtotals

    @lazyproperty
    def _columns_dimension(self):
        """The `Dimension` object representing column elements in this matrix."""
        return self._dimensions[1]

    @lazyproperty
    def _cube_result_matrix(self):
        """BaseCubeResultMatrix subclass object appropriate to this cube-slice.

        This matrix object encapsulates cube-result array parsing and MR multi-value
        differences and provides a foundational set of second-order analysis measure and
        margin arrays.
        """
        return BaseCubeResultMatrix.factory(
            self._cube, self._dimensions, self._slice_idx
        )

    def _dimension_labels(self, dimension, order):
        """1D str ndarray of name for each vector of `dimension`.

        These are suitable for use as row/column headings; labels for subtotals appear
        in the sequence specified by `order`, such that labels are ordered to correspond
        with their respective data vector.
        """
        return np.array(
            [e.label for e in dimension.valid_elements]
            + [s.label for s in dimension.subtotals]
        )[order]

    @lazyproperty
    def _measures(self):
        """SecondOrderMeasures collection object for this cube-result."""
        return SecondOrderMeasures(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def _row_order(self):
        """1D np.int64 ndarray of signed int idx for each assembled row.

        Negative values represent inserted subtotal-row locations.
        """
        return _BaseOrderHelper.row_display_order(self._dimensions, self._measures)

    @lazyproperty
    def _row_subtotals(self):
        """Sequence of _Subtotal object for each inserted row."""
        return self._rows_dimension.subtotals

    @lazyproperty
    def _rows_dimension(self):
        """The `Dimension` object representing row elements in this matrix."""
        return self._dimensions[0]


class _BaseOrderHelper(object):
    """Base class for ordering helpers."""

    def __init__(self, dimensions, second_order_measures):
        self._dimensions = dimensions
        self._second_order_measures = second_order_measures

    @classmethod
    def column_display_order(cls, dimensions, second_order_measures):
        """1D np.int64 ndarray of signed int idx for each column of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        columns-dimension.
        """
        # --- This is essentially a factory method. There is no sort-columns-by-value
        # --- yet, and both explicit and payload ordering are handled by
        # --- _ColumnOrderHelper, so there's not much to this yet, just keeping
        # --- form consistent with `.row_display_order()` and we'll elaborate this when
        # --- we add sort-by-value to columns.
        return _ColumnOrderHelper(dimensions, second_order_measures)._display_order

    @classmethod
    def row_display_order(cls, dimensions, second_order_measures):
        """1D np.int64 ndarray of signed int idx for each row of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        order_spec = dimensions[0].order_spec
        HelperCls = {
            CM.OPPOSING_ELEMENT: _SortRowsByBaseColumnHelper,
            CM.OPPOSING_INSERTION: _SortRowsByInsertedColumnHelper,
            CM.MARGINAL: _SortRowsByMarginalHelper,
        }.get(order_spec.collation_method, _RowOrderHelper)

        return HelperCls(dimensions, second_order_measures)._display_order

    @lazyproperty
    def _display_order(self):
        """1D np.int64 ndarray of signed int idx for each vector of dimension.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        # --- Returning as np.array suits its intended purpose, which is to participate
        # --- in an np._ix() call. It works fine as a sequence too for any alternate
        # --- use. Specifying int type prevents failure when there are zero elements.
        if self._prune_subtotals:
            return np.array([idx for idx in self._order if idx >= 0], dtype=int)
        return np.array(self._order, dtype=int)

    @lazyproperty
    def _columns_dimension(self):
        """The `Dimension` object representing column elements in the matrix."""
        return self._dimensions[1]

    @lazyproperty
    def _empty_column_idxs(self):
        """tuple of int index for each column with (unweighted) N = 0.

        These columns are subject to pruning, depending on a user setting in the
        dimension.
        """
        return tuple(
            i
            for i, N in enumerate(self._second_order_measures.columns_pruning_base)
            if N == 0
        )

    @lazyproperty
    def _empty_row_idxs(self):
        """tuple of int index for each row with N = 0.

        These rows are subject to pruning, depending on a user setting in the dimension.
        """
        return tuple(
            i
            for i, N in enumerate(self._second_order_measures.rows_pruning_base)
            if N == 0
        )

    @lazyproperty
    def _measure(self):
        """Second-order measure object providing values for sort.

        This property is not used by some subclasses.
        """
        propname_by_measure = {
            M.COLUMN_BASE_UNWEIGHTED: "column_unweighted_bases",
            M.COLUMN_BASE_WEIGHTED: "column_weighted_bases",
            M.COLUMN_PERCENT: "column_proportions",
            M.MEAN: "means",
            M.ROW_BASE_UNWEIGHTED: "row_unweighted_bases",
            M.ROW_BASE_WEIGHTED: "row_weighted_bases",
            M.ROW_PERCENT: "row_proportions",
            M.TABLE_BASE_UNWEIGHTED: "table_unweighted_bases",
            M.TABLE_BASE_WEIGHTED: "table_weighted_bases",
            M.UNWEIGHTED_COUNT: "unweighted_counts",
            M.WEIGHTED_COUNT: "weighted_counts",
            M.Z_SCORE: "zscores",
        }
        measure = self._order_spec.measure
        measure_propname = propname_by_measure.get(measure)

        if measure_propname is None:
            raise NotImplementedError(
                "sort-by-value for measure '%s' is not yet supported" % measure
            )

        return getattr(self._second_order_measures, measure_propname)

    @lazyproperty
    def _order(self):
        """tuple of signed int idx for each sorted vector of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `._order`" % type(self).__name__
        )

    @lazyproperty
    def _order_spec(self):
        """_OrderSpec object for dimension being sorted.

        Provides access to ordering details like measure and sort-direction.
        """
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `._order_spec`" % type(self).__name__
        )

    @lazyproperty
    def _prune_subtotals(self):
        """True if subtotal vectors need to be pruned, False otherwise."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `._prune_subtotals`" % type(self).__name__
        )

    @lazyproperty
    def _rows_dimension(self):
        """The `Dimension` object representing row elements in the matrix."""
        return self._dimensions[0]


class _ColumnOrderHelper(_BaseOrderHelper):
    """Encapsulates the complexity of the various kinds of column ordering."""

    @lazyproperty
    def _order(self):
        """tuple of signed int idx for each column of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        CollatorCls = (
            ExplicitOrderCollator
            if self._order_spec.collation_method == CM.EXPLICIT_ORDER
            else PayloadOrderCollator
        )
        return CollatorCls.display_order(
            self._columns_dimension, self._empty_column_idxs
        )

    @lazyproperty
    def _order_spec(self):
        """_OrderSpec object for the columns dimension.

        Provides access to ordering details like measure and sort-direction.
        """
        return self._columns_dimension.order_spec

    @lazyproperty
    def _prune_subtotals(self):
        """True if subtotal columns need to be pruned, False otherwise.

        Subtotal-columns need to be pruned when all base-rows are pruned. Subtotal
        columns are only subject to pruning when row-pruning is specified in the
        request.
        """
        return (
            len(self._empty_row_idxs) == len(self._rows_dimension.element_ids)
            if self._rows_dimension.prune
            else False
        )


class _RowOrderHelper(_BaseOrderHelper):
    """Encapsulates the complexity of the various kinds of row ordering."""

    @lazyproperty
    def _order(self):
        """tuple of signed int idx for each row of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        CollatorCls = (
            ExplicitOrderCollator
            if self._order_spec.collation_method == CM.EXPLICIT_ORDER
            else PayloadOrderCollator
        )
        return CollatorCls.display_order(self._rows_dimension, self._empty_row_idxs)

    @lazyproperty
    def _order_spec(self):
        """_OrderSpec object for this row dimension.

        Provides access to ordering details like measure and sort-direction.
        """
        return self._rows_dimension.order_spec

    @lazyproperty
    def _prune_subtotals(self):
        """True if subtotal rows need to be pruned, False otherwise.

        Subtotal-rows need to be pruned when all base-columns are pruned. Subtotal rows
        only subject to pruning when column-pruning is specified in the request.
        """
        return (
            len(self._empty_column_idxs) == len(self._columns_dimension.element_ids)
            if self._columns_dimension.prune
            else False
        )


class _SortRowsByBaseColumnHelper(_RowOrderHelper):
    """Orders elements by the values of an opposing base (not a subtotal) vector.

    This would be like "order rows in descending order by value of 'Strongly Agree'
    column. An opposing-element ordering is only available on a matrix, because only
    a matrix dimension has an opposing dimension.
    """

    @lazyproperty
    def _column_idx(self):
        """int index of column whose values the sort is based on."""
        column_element_ids = self._columns_dimension.element_ids
        sort_column_id = self._order_spec.element_id
        return column_element_ids.index(sort_column_id)

    @lazyproperty
    def _element_values(self):
        """Sequence of body values that form the basis for sort order.

        There is one value per row and values appear in payload (dimension) element
        order. These are only the "base" values and do not include insertions.
        """
        measure_base_values = self._measure.blocks[0][0]
        return measure_base_values[:, self._column_idx]

    @lazyproperty
    def _order(self):
        """tuple of int element-idx specifying ordering of dimension elements."""
        return SortByValueCollator.display_order(
            self._rows_dimension,
            self._element_values,
            self._subtotal_values,
            self._empty_row_idxs,
        )

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal values that contribute to the sort basis.

        There is one value per row subtotal and values appear in payload (dimension)
        insertion order.
        """
        measure_subtotal_rows = self._measure.blocks[1][0]
        return measure_subtotal_rows[:, self._column_idx]


class _SortRowsByInsertedColumnHelper(_RowOrderHelper):
    """Orders rows by the values in an inserted column.

    This would be like "order rows in descending order by value of 'Top 3' subtotal
    column. An opposing-insertion ordering is only available on a matrix because only
    a matrix dimension has an opposing dimension.
    """

    @lazyproperty
    def _element_values(self):
        """Sequence of insertion "body" values that form the basis for sort order.

        There is one value per base row and values appear in payload order. These are
        only the "base" insertion values and do not include intersections.
        """
        insertion_base_values = self._measure.blocks[0][1]
        return insertion_base_values[:, self._insertion_idx]

    @lazyproperty
    def _insertion_idx(self):
        """int index of insertion whose values the sort is based on."""
        return self._columns_dimension.insertion_ids.index(
            self._order_spec.insertion_id
        )

    @lazyproperty
    def _order(self):
        """tuple of int element-idx specifying ordering of dimension elements."""
        return SortByValueCollator.display_order(
            self._rows_dimension,
            self._element_values,
            self._subtotal_values,
            self._empty_row_idxs,
        )

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal intersection values that contribute to sort basis.

        There is one value per row subtotal and values appear in payload (dimension)
        insertion order.
        """
        intersections = self._measure.blocks[1][1]
        return intersections[:, self._insertion_idx]


class _SortRowsByMarginalHelper(_RowOrderHelper):
    """Orders elements by the values of an opposing marginal vector.

    This would be like "order rows in descending order by value of rows scale mean.
    column. An opposing-marginal ordering is only available on a matrix, because only
    a matrix dimension has an opposing dimension.
    """

    @lazyproperty
    def _element_values(self):
        """Sequence of body values that form the basis for sort order.

        There is one value per row and values appear in payload (dimension) element
        order. These are only the "base" values and do not include insertions.
        """
        return self._marginal.blocks[0]

    @lazyproperty
    def _marginal(self):
        """Marginal object providing values for sort."""
        marginal = self._order_spec.marginal
        marginal_propname = {
            MARGINAL.SCALE_MEAN: "rows_scale_mean",
            MARGINAL.SCALE_MEAN_STDDEV: "rows_scale_mean_stddev",
            MARGINAL.SCALE_MEDIAN: "rows_scale_median",
        }.get(marginal)

        if marginal_propname is None:
            raise NotImplementedError(
                "sort-by-value for marginal '%s' is not yet supported" % marginal
            )

        return getattr(self._second_order_measures, marginal_propname)

    @lazyproperty
    def _order(self):
        """tuple of int element-idx specifying ordering of dimension elements."""
        return SortByValueCollator.display_order(
            self._rows_dimension,
            self._element_values,
            self._subtotal_values,
            self._empty_row_idxs,
        )

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal values that contribute to the sort basis.

        There is one value per row subtotal and values appear in payload (dimension)
        insertion order.
        """
        return self._marginal.blocks[1]
